#!/usr/bin/env python3
# SPDX-FileCopyrightText: 2024 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Script which cleans up libraries for the bpy module.

It scans actual dependencies of the bpy module and its dependencies and removes libraries that
are not needed. The libraries that are needed are ensured to be regular files (not a symbolic
link).

The goal is to prepare the bpy install directory for the use by the wheel packaging tool: since
wheels don't support symbolic links leaving them in the install directory will result in big
resulting file sizes.

Doing cleanup as a dedicated step allows to run all sort of regression tests before the wheel is
packed.

Usage:
  wheel_cleanup.py <path/to/installed/bpy/directory>
"""

__all__ = (
    "main",
)

import argparse
import re
import shutil
import subprocess
import sys
from pathlib import Path
from string import digits
from typing import Generator


# REGEX matching for libraries that might be seen in the libs directory, not directly referenced,
# but yet are still required for the proper operation of the `bpy` module.
KEEP_MATCHERS = (
    # libOpenImageDenoise.so loads core, device_cuda, etc libraries at runtime.
    re.compile("libOpenImageDenoise_.*"),
)


def print_banner(text: str) -> None:
    print("")
    print(text)
    print("=" * len(text))
    print("")


def print_stage(text: str) -> None:
    print("")
    print(text)
    print("-" * len(text))
    print("")


def get_direct_elf_dependencies(elf: Path) -> set[str]:
    """
    Get direct dependencies of the given library or executable in ELF format

    Uses ``readelf`` command and parses its output.
    """
    output = subprocess.check_output(("readelf", "-d", elf))
    deps = set()
    for line_bytes in output.splitlines():
        # Example of a line from the readelf command:
        # b' 0x0000000000000001 (NEEDED)             Shared library: [libgcc_s.so.1]'
        line = line_bytes.decode()
        if "(NEEDED)" not in line:
            continue
        if "Shared library:" not in line:
            continue
        lib_quoted = line.split("Shared library: ")[1]
        lib = lib_quoted.removeprefix("[").removesuffix("]")
        deps.add(lib)
    return deps


def name_strip_abi_suffix(name: str) -> str:
    """
    Strip any ABI suffix from the given file name

    For example: ``libfoo.so.1.2-3`` -> ``linfoo.so``.
    """

    while name[-1] in digits:
        new_name = name.rstrip(digits)
        if new_name[-1] in (".", "-"):
            name = new_name[:-1]
        else:
            break
    return name


def name_is_so(name: str) -> bool:
    """
    Return true if the given name is an .so library

    Ignores any possible ABI specification.
    This is purely lexicographical operation.
    """
    clean_name = name_strip_abi_suffix(name)
    return clean_name.endswith(".so")


def iter_so_in_dir(parent_dir: Path) -> Generator[Path, None, None]:
    """
    Iterate .so files (with ABI variants) in the given directory

    The file are yielded from this generator.
    """
    for lib_filepath in parent_dir.iterdir():
        if not lib_filepath.is_file():
            continue
        if not name_is_so(lib_filepath.name):
            continue
        yield lib_filepath


def resolve_symlink(filepath: Path) -> Path:
    """
    Resolve symbolic link

    Recursively follows the resolution.

    NOTE: Does not support cyclic symbolic links
    """

    if not filepath.is_symlink():
        return filepath

    link = Path(filepath.readlink())
    return resolve_symlink(filepath.parent / link)


def make_real(filepath: Path) -> None:
    """
    Make the given file real by resolving symbolic link
    """

    print(f"Making {filepath} real")
    if not filepath.is_symlink():
        print(f"{filepath} is not a link")
        return

    resolved = resolve_symlink(filepath)

    print(f"Resolved to {resolved}")

    filepath.unlink()
    shutil.copy2(resolved, filepath)


def library_matches_keep_pattern(lib_filepath: Path) -> bool:
    """
    Returns true if the library matches any pattern that requires it to be kept
    """
    for matcher in KEEP_MATCHERS:
        if matcher.match(lib_filepath.name):
            return True
    return False


def cleanup_linux(bpy_dir: Path) -> None:
    print_stage("Gathering dependencies")
    deps = get_direct_elf_dependencies(bpy_dir / "__init__.so")
    print(f"- __init__.so depends on {deps}")
    # TODO(sergey): Can do something smarter like actual recursive dependency tracing.
    for lib_filepath in iter_so_in_dir(bpy_dir / "lib"):
        lib_deps = deps.union(get_direct_elf_dependencies(lib_filepath))
        print(f"- {lib_filepath} depends on {lib_deps}")
        deps = deps.union(lib_deps)

    print_stage("Cleaning up")
    deps_to_remove = []
    for lib_filepath in iter_so_in_dir(bpy_dir / "lib"):
        if lib_filepath.name in deps:
            print(f"Keeping dependency {lib_filepath}")
            make_real(lib_filepath)
            continue

        if library_matches_keep_pattern(lib_filepath):
            print(f"Keeping dependency {lib_filepath} as per static rules")
            continue

        deps_to_remove.append(lib_filepath)
        print(f"Will remove unused dependency {lib_filepath}")

    print_stage("Removing unused dependencies")
    for dep_to_remove in deps_to_remove:
        print(f"Removing unused dependency {dep_to_remove}")
        dep_to_remove.unlink()


def main() -> None:
    print_banner("BPY module libraries cleaner")

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument("bpy_dir", type=Path, help="Path to the bpy directory to cleanup")
    args = parser.parse_args()

    if sys.platform == "linux":
        cleanup_linux(args.bpy_dir)
        return

    # Windows and macOS do not use symlinks for libraries, so no need to figure out which copies of
    # the same library with different ABI can be removed.
    print("The wheel cleanup script is only intended to be used on Linux")


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2011-2022 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

# Simple script to enable all add-ons, and disable.

"""
./blender.bin --background --factory-startup --python tests/python/bl_load_addons.py
"""
__all__ = (
    "main",
)

import bpy
import addon_utils

import sys
import importlib

EXCLUDE_DIRECTORIES = (
    bpy.utils.user_resource('SCRIPTS'),
    *addon_utils.paths()[1:],
)
EXCLUDE_ADDONS = set()


def _init_addon_exclusions():

    # In case we built without cycles.
    if not bpy.app.build_options.cycles:
        EXCLUDE_ADDONS.add("cycles")

    # In case we built without freestyle.
    if not bpy.app.build_options.freestyle:
        EXCLUDE_ADDONS.add("render_freestyle_svg")

    if not bpy.app.build_options.xr_openxr:
        EXCLUDE_ADDONS.add("viewport_vr_preview")


def addon_modules_sorted():
    # Pass in an empty module cache to prevent `addon_utils` local module cache being manipulated.
    modules = [
        mod for mod in addon_utils.modules(module_cache={})
        if not (mod.__file__.startswith(EXCLUDE_DIRECTORIES))
        if not (mod.__name__ in EXCLUDE_ADDONS)
    ]
    modules.sort(key=lambda mod: mod.__name__)
    return modules


def disable_addons():
    # First disable all.
    addons = bpy.context.preferences.addons
    for mod_name in list(addons.keys()):
        addon_utils.disable(mod_name, default_set=True)
    assert bool(addons) is False


def test_load_addons(prefs):
    modules = addon_modules_sorted()

    disable_addons()

    addons = prefs.addons

    addons_fail = []

    for mod in modules:
        mod_name = mod.__name__
        print("\tenabling:", mod_name)
        addon_utils.enable(mod_name, default_set=True)
        if (mod_name not in addons) and (mod_name not in EXCLUDE_ADDONS):
            addons_fail.append(mod_name)

    if addons_fail:
        print("addons failed to load (%d):" % len(addons_fail))
        for mod_name in addons_fail:
            print("    %s" % mod_name)
    else:
        print("addons all loaded without errors!")
    print("")


def reload_addons(prefs, do_reload=True, do_reverse=True):
    modules = addon_modules_sorted()
    addons = prefs.addons

    disable_addons()

    # Run twice each time.
    for _ in (0, 1):
        for mod in modules:
            mod_name = mod.__name__
            print("\tenabling:", mod_name)
            addon_utils.enable(mod_name, default_set=True)
            assert mod_name in addons

        for mod in modules:
            mod_name = mod.__name__
            print("\tdisabling:", mod_name)
            addon_utils.disable(mod_name, default_set=True)
            assert not (mod_name in addons)

            # Now test reloading.
            if do_reload:
                sys.modules[mod_name] = importlib.reload(sys.modules[mod_name])

        if do_reverse:
            # In case order matters when it shouldn't.
            modules.reverse()


def main():
    from bpy import context
    # Remove repositories since these point to the users home directory by default.
    prefs = context.preferences
    extension_repos = prefs.extensions.repos
    while extension_repos:
        extension_repos.remove(extension_repos[0])

    _init_addon_exclusions()

    # First load add-ons, print a list of all add-ons that fail.
    test_load_addons(prefs)

    reload_addons(prefs, do_reload=False, do_reverse=False)
    reload_addons(prefs, do_reload=False, do_reverse=True)
    reload_addons(prefs, do_reload=True, do_reverse=True)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2020-2023 Blender Authors
#
# SPDX-License-Identifier: Apache-2.0

# NOTE: See also `bl_pyapi_prop_array.py` for the `Vector` bpy.props similar tests,
# and `bl_pyapi_idprop.py` for some deeper testing of the consistency between
# the underlying IDProperty storage, and the property data exposed in Python.

# ./blender.bin --background --python tests/python/bl_pyapi_prop.py -- --verbose
import bpy
from bpy.props import (
    BoolProperty,
    IntProperty,
    FloatProperty,
    EnumProperty,
    StringProperty,
    PointerProperty,
    CollectionProperty,
)

import unittest
import functools

id_inst = bpy.context.scene
id_type = bpy.types.Scene


# -----------------------------------------------------------------------------
# Utility Types

class TestPropertyGroup(bpy.types.PropertyGroup):
    test_prop: IntProperty()


# -----------------------------------------------------------------------------
# Tests

class TestPropNumerical(unittest.TestCase):
    default_value = 0
    custom_value = 1
    min_value = -1
    max_value = 5

    def setUp(self):
        id_type.test_bool = BoolProperty(default=bool(self.default_value))
        id_type.test_int = IntProperty(
            default=int(self.default_value),
            min=int(self.min_value),
            max=int(self.max_value),
        )
        id_type.test_float = FloatProperty(
            default=float(self.default_value),
            min=float(self.min_value),
            max=float(self.max_value),
        )

        self.test_bool_storage = bool(self.custom_value)
        self.test_int_storage = int(self.custom_value)
        self.test_float_storage = float(self.custom_value)

        def bool_set_(s, v):
            self.test_bool_storage = v

        def int_set_(s, v):
            self.test_int_storage = v

        def float_set_(s, v):
            self.test_float_storage = v

        id_type.test_bool_getset = BoolProperty(
            default=bool(self.default_value),
            get=lambda s: self.test_bool_storage,
            set=bool_set_,
        )
        id_type.test_int_getset = IntProperty(
            default=int(self.default_value),
            min=int(self.min_value),
            max=int(self.max_value),
            get=lambda s: self.test_int_storage,
            set=int_set_,
        )
        id_type.test_float_getset = FloatProperty(
            default=float(self.default_value),
            min=float(self.min_value),
            max=float(self.max_value),
            get=lambda s: self.test_float_storage,
            set=float_set_,
        )

        id_type.test_bool_transform = BoolProperty(
            default=bool(self.default_value),
            get_transform=lambda s, c_v, isset: c_v,
            set_transform=lambda s, n_v, c_v, isset: n_v,
        )
        id_type.test_int_transform = IntProperty(
            default=int(self.default_value),
            min=int(self.min_value),
            max=int(self.max_value),
            get_transform=lambda s, c_v, isset: c_v,
            set_transform=lambda s, n_v, c_v, isset: n_v,
        )
        id_type.test_float_transform = FloatProperty(
            default=float(self.default_value),
            min=float(self.min_value),
            max=float(self.max_value),
            get_transform=lambda s, c_v, isset: c_v,
            set_transform=lambda s, n_v, c_v, isset: n_v,
        )

        id_type.test_bool_getset_transform = BoolProperty(
            default=bool(self.default_value),
            get=lambda s: self.test_bool_storage,
            set=bool_set_,
            get_transform=lambda s, c_v, isset: c_v,
            set_transform=lambda s, n_v, c_v, isset: n_v,
        )
        id_type.test_int_getset_transform = IntProperty(
            default=int(self.default_value),
            min=int(self.min_value),
            max=int(self.max_value),
            get=lambda s: self.test_int_storage,
            set=int_set_,
            get_transform=lambda s, c_v, isset: c_v,
            set_transform=lambda s, n_v, c_v, isset: n_v,
        )
        id_type.test_float_getset_transform = FloatProperty(
            default=float(self.default_value),
            min=float(self.min_value),
            max=float(self.max_value),
            get=lambda s: self.test_float_storage,
            set=float_set_,
            get_transform=lambda s, c_v, isset: c_v,
            set_transform=lambda s, n_v, c_v, isset: n_v,
        )

    def tearDown(self):
        del id_type.test_float
        del id_type.test_int
        del id_type.test_bool

        del id_type.test_float_getset
        del id_type.test_int_getset
        del id_type.test_bool_getset

        del id_type.test_float_transform
        del id_type.test_int_transform
        del id_type.test_bool_transform

        del id_type.test_float_getset_transform
        del id_type.test_int_getset_transform
        del id_type.test_bool_getset_transform

    def do_min_max_expect_success(self, prop_name, py_type):
        # This property is expected to properly clamp set values within required range.
        setattr(id_inst, prop_name, py_type(self.min_value - 1))
        self.assertEqual(getattr(id_inst, prop_name), py_type(self.min_value))
        setattr(id_inst, prop_name, py_type(self.max_value + 1))
        self.assertEqual(getattr(id_inst, prop_name), py_type(self.max_value))

    def do_min_max_expect_failure(self, prop_name, py_type):
        # This property is not expected to properly clamp set values within required range.
        # This happens when using custom setters.
        setattr(id_inst, prop_name, py_type(self.min_value - 1))
        self.assertNotEqual(getattr(id_inst, prop_name), py_type(self.min_value))
        setattr(id_inst, prop_name, py_type(self.max_value + 1))
        self.assertNotEqual(getattr(id_inst, prop_name), py_type(self.max_value))

    def do_test_access(self, prop_name, py_type, expected_value, do_min_max=None):
        v = getattr(id_inst, prop_name)
        self.assertIsInstance(v, py_type)
        self.assertEqual(v, expected_value)
        setattr(id_inst, prop_name, v)
        self.assertEqual(getattr(id_inst, prop_name), expected_value)
        setattr(id_inst, prop_name, py_type(self.custom_value))
        self.assertEqual(getattr(id_inst, prop_name), py_type(self.custom_value))
        if do_min_max:
            do_min_max(prop_name, py_type)

    def test_access_bool(self):
        self.do_test_access("test_bool", bool, bool(self.default_value))

    def test_access_int(self):
        self.do_test_access(
            "test_int",
            int,
            int(self.default_value),
            do_min_max=self.do_min_max_expect_success,
        )

    def test_access_float(self):
        self.do_test_access(
            "test_float",
            float,
            float(self.default_value),
            do_min_max=self.do_min_max_expect_success,
        )

    def test_access_bool_getset(self):
        self.do_test_access("test_bool_getset", bool, bool(self.custom_value))

    def test_access_int_getset(self):
        self.do_test_access(
            "test_int_getset",
            int,
            int(self.custom_value),
            do_min_max=self.do_min_max_expect_failure,
        )

    def test_access_float_getset(self):
        self.do_test_access(
            "test_float_getset",
            float,
            float(self.custom_value),
            do_min_max=self.do_min_max_expect_failure,
        )

    def test_access_bool_transform(self):
        self.do_test_access("test_bool_transform", bool, bool(self.default_value))

    def test_access_int_transform(self):
        self.do_test_access(
            "test_int_transform",
            int,
            int(self.default_value),
            do_min_max=self.do_min_max_expect_success,
        )

    def test_access_float_transform(self):
        self.do_test_access(
            "test_float_transform",
            float,
            float(self.default_value),
            do_min_max=self.do_min_max_expect_success,
        )

    def test_access_bool_getset_transform(self):
        self.do_test_access("test_bool_getset_transform", bool, bool(self.custom_value))

    def test_access_int_getset_transform(self):
        self.do_test_access(
            "test_int_getset_transform",
            int,
            int(self.custom_value),
            do_min_max=self.do_min_max_expect_failure,
        )

    def test_access_float_getset_transform(self):
        self.do_test_access(
            "test_float_getset_transform",
            float,
            float(self.custom_value),
            do_min_max=self.do_min_max_expect_failure,
        )

    # TODO: Add expected failure cases (e.g. handling of out-of range values).


class TestPropString(unittest.TestCase):
    default_value = ""
    custom_value = "Blender"

    def setUp(self):
        id_type.test_string = StringProperty(default=self.default_value)

        self.test_string_storage = self.custom_value

        def set_(s, v):
            self.test_string_storage = v
        id_type.test_string_getset = StringProperty(
            default=self.default_value,
            get=lambda s: self.test_string_storage,
            set=set_,
        )

    def tearDown(self):
        del id_type.test_string
        del id_type.test_string_getset

    def do_test_access(self, prop_name, py_type, expected_value):
        v = getattr(id_inst, prop_name)
        self.assertIsInstance(v, py_type)
        self.assertEqual(v, expected_value)
        setattr(id_inst, prop_name, v)

    def test_access_string(self):
        self.do_test_access("test_string", str, self.default_value)

    def test_access_string_getset(self):
        self.do_test_access("test_string_getset", str, self.custom_value)

    # TODO: Add expected failure cases (e.g. handling of too long values, invalid utf8 sequences, etc.).


class TestPropByteString(unittest.TestCase):
    default_value = b""
    custom_value = b"Blender"
    custom_nullbyte_value = b"Blen\0der"

    def setUp(self):
        id_type.test_byte_string = StringProperty(default=self.default_value.decode(), subtype="BYTE_STRING")

        self.test_byte_string_storage = self.custom_value

        def set_(s, v):
            self.test_byte_string_storage = v
        id_type.test_byte_string_getset = StringProperty(
            default=self.default_value.decode(),
            subtype="BYTE_STRING",
            get=lambda s: self.test_byte_string_storage,
            set=set_,
        )

    def tearDown(self):
        del id_type.test_byte_string

    def do_test_access(self, prop_name, py_type, expected_value):
        v = getattr(id_inst, prop_name)
        self.assertIsInstance(v, py_type)
        self.assertEqual(v, expected_value)
        # Test with nullbyte in the value.
        setattr(id_inst, prop_name, self.custom_nullbyte_value)
        v = getattr(id_inst, prop_name)
        self.assertIsInstance(v, py_type)
        self.assertEqual(v, self.custom_nullbyte_value)

    def test_access_byte_string(self):
        self.do_test_access("test_byte_string", bytes, self.default_value)

    # TODO: get/set is known essentially broken (aka unsupported) with byte strings properties currently.
    # See #150755, and comments in `bpy_props.cc`:
    # `bpy_prop_string_get_transform_locked_fn`, `bpy_prop_string_set_fn`, etc.
    # def test_access_byte_string_getset(self):
    #     self.do_test_access("test_byte_string_getset", bytes, self.custom_value)

    # TODO: Add expected failure cases (e.g. handling of too long values, etc.).


class TestPropEnum(unittest.TestCase):
    # FIXME: Auto-generated enum values do not play well with partially specifying some values.
    # This won't work, generating (1, 1, 2, 16):
    #   enum_items = (("1", "1", "", 1), ("2", "2", ""), ("3", "3", ""), ("16", "16", "", 16),)
    enum_items = (("1", "1", ""), ("2", "2", ""), ("3", "3", ""), ("16", "16", "", 16),)
    enum_expected_values = {"1": 0, "2": 1, "3": 2, "16": 16}
    enum_expected_bitflag_values = {"1": 2**0, "2": 2**1, "3": 2**2, "16": 2**4}

    default_value = "1"
    custom_value = "2"
    default_bitflag_value = {"1", "3"}
    custom_bitflag_value = {"16", "3"}

    def setUp(self):
        id_type.test_enum = EnumProperty(items=self.enum_items, default=self.default_value)
        id_type.test_enum_bitflag = EnumProperty(
            items=self.enum_items,
            default=self.default_bitflag_value,
            options={"ENUM_FLAG"},
        )

        self.test_enum_storage = self.enum_expected_values[self.custom_value]
        self.test_enum_bitflag_storage = functools.reduce(
            lambda a, b: a | b,
            (self.enum_expected_bitflag_values[bf] for bf in self.custom_bitflag_value))

        def enum_set_(s, v):
            self.test_enum_storage = v

        def bitflag_set_(s, v):
            self.test_enum_bitflag_storage = v

        id_type.test_enum_getset = EnumProperty(
            items=self.enum_items,
            default=self.default_value,
            get=lambda s: self.test_enum_storage,
            set=enum_set_,
        )

        id_type.test_enum_bitflag_getset = EnumProperty(
            items=self.enum_items,
            default=self.default_bitflag_value,
            options={"ENUM_FLAG"},
            get=lambda s: self.test_enum_bitflag_storage,
            set=bitflag_set_,
        )

        id_type.test_enum_transform = EnumProperty(
            items=self.enum_items,
            default=self.default_value,
            get_transform=lambda s, c_v, isset: c_v,
            set_transform=lambda s, n_v, c_v, isset: n_v,
        )

        id_type.test_enum_bitflag_transform = EnumProperty(
            items=self.enum_items,
            default=self.default_bitflag_value,
            options={"ENUM_FLAG"},
            get_transform=lambda s, c_v, isset: c_v,
            set_transform=lambda s, n_v, c_v, isset: n_v,
        )

        id_type.test_enum_getset_transform = EnumProperty(
            items=self.enum_items,
            default=self.default_value,
            get=lambda s: self.test_enum_storage,
            set=enum_set_,
            get_transform=lambda s, c_v, isset: c_v,
            set_transform=lambda s, n_v, c_v, isset: n_v,
        )

        id_type.test_enum_bitflag_getset_transform = EnumProperty(
            items=self.enum_items,
            default=self.default_bitflag_value,
            options={"ENUM_FLAG"},
            get_transform=lambda s, c_v, isset: c_v,
            set_transform=lambda s, n_v, c_v, isset: n_v,
            get=lambda s: self.test_enum_bitflag_storage,
            set=bitflag_set_,
        )

    def tearDown(self):
        del id_type.test_enum
        del id_type.test_enum_bitflag
        del id_type.test_enum_getset
        del id_type.test_enum_bitflag_getset
        del id_type.test_enum_transform
        del id_type.test_enum_bitflag_transform
        del id_type.test_enum_getset_transform
        del id_type.test_enum_bitflag_getset_transform

    # Test expected generated values for enum items.
    def do_test_enum_values(self, prop_name, expected_item_values):
        enum_items = id_inst.bl_rna.properties[prop_name].enum_items
        self.assertEqual(len(expected_item_values), len(enum_items))
        for (expected_identifier, expected_value), item in zip(expected_item_values.items(), enum_items):
            self.assertEqual(expected_identifier, item.identifier)
            self.assertEqual(expected_value, item.value)

    def test_enum_item_values(self):
        self.do_test_enum_values("test_enum", self.enum_expected_values)

    def test_enum_bitflag_item_values(self):
        self.do_test_enum_values("test_enum_bitflag", self.enum_expected_bitflag_values)

    def test_enum_getset_item_values(self):
        self.do_test_enum_values("test_enum_getset", self.enum_expected_values)

    def test_enum_bitflag_getset_item_values(self):
        self.do_test_enum_values("test_enum_bitflag_getset", self.enum_expected_bitflag_values)

    def test_enum_transform_item_values(self):
        self.do_test_enum_values("test_enum_transform", self.enum_expected_values)

    def test_enum_bitflag_transform_item_values(self):
        self.do_test_enum_values("test_enum_bitflag_transform", self.enum_expected_bitflag_values)

    def test_enum_getset_transform_item_values(self):
        self.do_test_enum_values("test_enum_getset_transform", self.enum_expected_values)

    def test_enum_bitflag_getset_transform_item_values(self):
        self.do_test_enum_values("test_enum_bitflag_getset_transform", self.enum_expected_bitflag_values)

    # Test basic access to enum values.
    def do_test_access(self, prop_name, py_type, expected_value):
        v = getattr(id_inst, prop_name)
        self.assertIsInstance(v, py_type)
        self.assertEqual(v, expected_value)
        setattr(id_inst, prop_name, v)

    def test_access_enum(self):
        self.do_test_access("test_enum", str, self.default_value)

    def test_access_enum_bitflag(self):
        self.do_test_access("test_enum_bitflag", set, self.default_bitflag_value)

    def test_access_enum_getset(self):
        self.do_test_access("test_enum_getset", str, self.custom_value)

    def test_access_enum_bitflag_getset(self):
        self.do_test_access("test_enum_bitflag_getset", set, self.custom_bitflag_value)

    def test_access_enum_transform(self):
        self.do_test_access("test_enum_transform", str, self.default_value)

    def test_access_enum_bitflag_transform(self):
        self.do_test_access("test_enum_bitflag_transform", set, self.default_bitflag_value)

    def test_access_enum_getset_transform(self):
        self.do_test_access("test_enum_getset_transform", str, self.custom_value)

    def test_access_enum_bitflag_getset_transform(self):
        self.do_test_access("test_enum_bitflag_getset_transform", set, self.custom_bitflag_value)

    # TODO: Add expected failure cases (e.g. handling of invalid items identifiers).


class TestPropCollectionAndPointer(unittest.TestCase):

    def setUp(self):
        bpy.utils.register_class(TestPropertyGroup)

        id_type.test_pointer = PointerProperty(type=TestPropertyGroup)
        id_type.test_pointer_ID = PointerProperty(type=bpy.types.ID)
        id_type.test_pointer_ID_poll = PointerProperty(type=bpy.types.ID, poll=lambda s, v: v.id_type == 'OBJECT')
        id_type.test_collection = CollectionProperty(type=TestPropertyGroup)

    def tearDown(self):
        del id_type.test_pointer
        del id_type.test_pointer_ID
        del id_type.test_pointer_ID_poll
        del id_type.test_collection

        bpy.utils.unregister_class(TestPropertyGroup)

    def test_access_pointer(self):
        v = id_inst.test_pointer
        self.assertIsInstance(v, TestPropertyGroup)
        self.assertTrue(hasattr(v, "test_prop"))
        self.assertEqual(v.test_prop, 0)
        v.test_prop = 42
        self.assertEqual(id_inst.test_pointer.test_prop, 42)

    def test_access_pointer_ID(self):
        self.assertEqual(id_inst.test_pointer_ID, None)

        # Non-refcounting ID type
        win_man = bpy.data.window_managers[0]
        win_man_users = win_man.users
        id_inst.test_pointer_ID = win_man
        self.assertEqual(id_inst.test_pointer_ID, win_man)
        self.assertEqual(win_man.users, win_man_users)
        id_inst.test_pointer_ID = None
        self.assertEqual(id_inst.test_pointer_ID, None)
        self.assertEqual(win_man.users, win_man_users)

        # Refcounting ID type
        ma = bpy.data.materials[0]
        ma_users = ma.users
        id_inst.test_pointer_ID = ma
        self.assertEqual(id_inst.test_pointer_ID, ma)
        self.assertEqual(win_man.users, ma_users + 1)
        id_inst.test_pointer_ID = None
        self.assertEqual(id_inst.test_pointer_ID, None)
        self.assertEqual(ma.users, ma_users)

    def test_access_pointer_ID_poll(self):
        # Poll callback is only used for UI, in scripts it's still possible to assign an 'invalid' ID.
        self.assertEqual(id_inst.test_pointer_ID_poll, None)
        win_man = bpy.data.window_managers[0]
        id_inst.test_pointer_ID_poll = win_man
        self.assertEqual(id_inst.test_pointer_ID_poll, win_man)
        id_inst.test_pointer_ID_poll = None
        self.assertEqual(id_inst.test_pointer_ID_poll, None)

    def test_access_collection(self):
        self.assertEqual(len(id_inst.test_collection), 0)

        test_item = id_inst.test_collection.add()
        self.assertEqual(len(id_inst.test_collection), 1)
        self.assertIsInstance(test_item, TestPropertyGroup)
        self.assertTrue(hasattr(test_item, "test_prop"))
        self.assertEqual(test_item.test_prop, 0)
        self.assertEqual(id_inst.test_collection[0], test_item)
        test_item.test_prop = 42
        self.assertEqual(test_item.test_prop, 42)
        self.assertEqual(id_inst.test_collection[0].test_prop, 42)

        test_item_2 = id_inst.test_collection.add()
        test_item_3 = id_inst.test_collection.add()
        test_item_3.test_prop = 24
        self.assertEqual(len(id_inst.test_collection), 3)
        self.assertEqual(id_inst.test_collection[0], test_item)
        self.assertEqual(id_inst.test_collection[1], test_item_2)
        self.assertEqual(id_inst.test_collection[2], test_item_3)

        id_inst.test_collection.remove(1)
        self.assertEqual(len(id_inst.test_collection), 2)
        self.assertEqual(id_inst.test_collection[0], test_item)
        # Removing the second item re-allocates the third one, so no equality anymore.
        self.assertNotEqual(id_inst.test_collection[1], test_item_3)
        self.assertEqual(id_inst.test_collection[1].test_prop, 24)

    # TODO: Add expected failure cases (e.g. assigning propertygroup to a Pointer property, etc.).


if __name__ == '__main__':
    import sys
    sys.argv = [__file__] + (sys.argv[sys.argv.index("--") + 1:] if "--" in sys.argv else [])
    unittest.main()



"""
2D Image
--------

To use this example you have to provide an image that should be displayed.
"""
import bpy
import gpu
from gpu_extras.batch import batch_for_shader

IMAGE_NAME = "Untitled"
image = bpy.data.images[IMAGE_NAME]
texture = gpu.texture.from_image(image)

shader = gpu.shader.from_builtin('IMAGE_SCENE_LINEAR_TO_REC709_SRGB')
batch = batch_for_shader(
    shader, 'TRI_FAN',
    {
        "pos": ((100, 100), (200, 100), (200, 200), (100, 200)),
        "texCoord": ((0, 0), (1, 0), (1, 1), (0, 1)),
    },
)


def draw():
    shader.bind()
    shader.uniform_sampler("image", texture)
    batch.draw(shader)


bpy.types.SpaceView3D.draw_handler_add(draw, (), 'WINDOW', 'POST_PIXEL')

"""
3D Image
--------

Similar to the 2D Image shader, but works with 3D positions for the image vertices.
To use this example you have to provide an image that should be displayed.
"""
import bpy
import gpu
from gpu_extras.batch import batch_for_shader

IMAGE_NAME = "Untitled"
image = bpy.data.images[IMAGE_NAME]
texture = gpu.texture.from_image(image)

shader = gpu.shader.from_builtin('IMAGE_SCENE_LINEAR_TO_REC709_SRGB')
batch = batch_for_shader(
    shader, 'TRIS',
    {
        "pos": ((0, 0, 0), (0, 1, 1), (1, 1, 1), (1, 1, 1), (1, 0, 0), (0, 0, 0)),
        "texCoord": ((0, 0), (0, 1), (1, 1), (1, 1), (1, 0), (0, 0)),
    },
)


def draw():
    shader.uniform_sampler("image", texture)
    batch.draw(shader)


bpy.types.SpaceView3D.draw_handler_add(draw, (), 'WINDOW', 'POST_VIEW')



# SPDX-FileCopyrightText: 2021-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

import pathlib
import sys
import unittest
import tempfile

import bpy

args = None


class AbstractNodeGroupInterfaceTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.testdir = args.testdir
        cls._tempdir = tempfile.TemporaryDirectory()
        cls.tempdir = pathlib.Path(cls._tempdir.name)

    def setUp(self):
        self.assertTrue(self.testdir.exists(),
                        'Test dir {0} should exist'.format(self.testdir))

        # Make sure we always start with a known-empty file.
        bpy.ops.wm.open_mainfile(filepath=str(self.testdir / "empty.blend"))

    def tearDown(self):
        self._tempdir.cleanup()


class NodeGroupInterfaceTests:
    tree_type = None
    group_node_type = None
    # Tree instance where node groups can be added
    main_tree = None

    def make_group(self):
        tree = bpy.data.node_groups.new("test", self.tree_type)
        return tree

    def make_instance(self, tree):
        group_node = self.main_tree.nodes.new(self.group_node_type)
        group_node.node_tree = tree
        return group_node

    def make_group_and_instance(self):
        tree = self.make_group()
        group_node = self.make_instance(tree)
        return tree, group_node

    # Utility method for generating a non-zero default value.
    @staticmethod
    def make_default_socket_value(socket_type):
        if (socket_type == "NodeSocketBool"):
            return True
        elif (socket_type == "NodeSocketColor"):
            return (.5, 1.0, .3, .7)
        elif (socket_type == "NodeSocketFloat"):
            return 1.23
        elif (socket_type == "NodeSocketImage"):
            return bpy.data.images.new("test", 4, 4)
        elif (socket_type == "NodeSocketInt"):
            return -6
        elif (socket_type == "NodeSocketMaterial"):
            return bpy.data.materials.new("test")
        elif (socket_type == "NodeSocketObject"):
            return bpy.data.objects.new("test", bpy.data.meshes.new("test"))
        elif (socket_type == "NodeSocketRotation"):
            return (0.3, 5.0, -42)
        elif (socket_type == "NodeSocketString"):
            return "Hello World!"
        elif (socket_type == "NodeSocketVector"):
            return (4.0, -1.0, 0.0)

    # Utility method returning a comparator for socket values.
    # Not all socket value types are trivially comparable, e.g. colors.
    @staticmethod
    def make_socket_value_comparator(socket_type):
        def cmp_default(test, value, expected):
            test.assertEqual(value, expected, f"Value {value} does not match expected value {expected}")

        def cmp_array(test, value, expected):
            test.assertSequenceEqual(value[:], expected[:], f"Value {value} does not match expected value {expected}")

        if (socket_type in {"NodeSocketBool",
                            "NodeSocketFloat",
                            "NodeSocketImage",
                            "NodeSocketInt",
                            "NodeSocketMaterial",
                            "NodeSocketObject",
                            "NodeSocketRotation",
                            "NodeSocketString"}):
            return cmp_default
        elif (socket_type in {"NodeSocketColor",
                              "NodeSocketVector"}):
            return cmp_array

    def test_empty_nodegroup(self):
        tree, group_node = self.make_group_and_instance()

        self.assertFalse(tree.interface.items_tree, "Interface not empty")
        self.assertFalse(group_node.inputs)
        self.assertFalse(group_node.outputs)

    def do_test_invalid_socket_type(self, socket_type):
        tree = self.make_group()

        with self.assertRaises(TypeError):
            in0 = tree.interface.new_socket("Input 0", socket_type=socket_type, in_out='INPUT')
            self.assertIsNone(in0, f"Socket created for invalid type {socket_type}")
        with self.assertRaises(TypeError):
            out0 = tree.interface.new_socket("Output 0", socket_type=socket_type, in_out='OUTPUT')
            self.assertIsNone(out0, f"Socket created for invalid type {socket_type}")

    def do_test_sockets_in_out(self, socket_type):
        tree, group_node = self.make_group_and_instance()

        out0 = tree.interface.new_socket("Output 0", socket_type=socket_type, in_out='OUTPUT')
        self.assertIsNotNone(out0, f"Could not create socket of type {socket_type}")

        in0 = tree.interface.new_socket("Input 0", socket_type=socket_type, in_out='INPUT')
        self.assertIsNotNone(in0, f"Could not create socket of type {socket_type}")

        in1 = tree.interface.new_socket("Input 1", socket_type=socket_type, in_out='INPUT')
        self.assertIsNotNone(in1, f"Could not create socket of type {socket_type}")

        out1 = tree.interface.new_socket("Output 1", socket_type=socket_type, in_out='OUTPUT')
        self.assertIsNotNone(out1, f"Could not create socket of type {socket_type}")

        self.assertSequenceEqual([(s.name, s.bl_idname) for s in group_node.inputs], [
            ("Input 0", socket_type),
            ("Input 1", socket_type),
        ])
        self.assertSequenceEqual([(s.name, s.bl_idname) for s in group_node.outputs], [
            ("Output 0", socket_type),
            ("Output 1", socket_type),
        ])

    def do_test_user_count(self, value, expected_users):
        if (isinstance(value, bpy.types.ID)):
            self.assertEqual(
                value.users,
                expected_users,
                f"Socket default value has user count {value.users}, expected {expected_users}")

    def do_test_socket_type(self, socket_type):
        default_value = self.make_default_socket_value(socket_type)
        compare_value = self.make_socket_value_comparator(socket_type)

        # Create the tree first, add sockets, then create a group instance.
        # That way the new instance should reflect the expected default values.
        tree = self.make_group()

        in0 = tree.interface.new_socket("Input 0", socket_type=socket_type, in_out='INPUT')
        if default_value is not None:
            in0.default_value = default_value
        out0 = tree.interface.new_socket("Output 0", socket_type=socket_type, in_out='OUTPUT')
        self.assertIsNotNone(in0, f"Could not create socket of type {socket_type}")
        self.assertIsNotNone(out0, f"Could not create socket of type {socket_type}")

        # Now make a node group instance to check default values.
        group_node = self.make_instance(tree)
        if compare_value:
            compare_value(self, group_node.inputs[0].default_value, in0.default_value)

        # Test ID user count after assigning.
        if (hasattr(in0, "default_value")):
            # The default value is stored in both the interface and node, it should have 2 users now.
            self.do_test_user_count(in0.default_value, 2)

        # Copy sockets
        in1 = tree.interface.copy(in0)
        out1 = tree.interface.copy(out0)
        self.assertIsNotNone(in1, "Could not copy socket")
        self.assertIsNotNone(out1, "Could not copy socket")
        # User count on default values should increment by 2 after copy,
        # one user for the interface and one for the group node instance.
        if (hasattr(in1, "default_value")):
            self.do_test_user_count(in1.default_value, 4)

    # Classic outputs..inputs socket layout
    def do_test_items_order_classic(self, socket_type):
        tree, group_node = self.make_group_and_instance()

        tree.interface.new_socket("Output 0", socket_type=socket_type, in_out='OUTPUT')
        tree.interface.new_socket("Input 0", socket_type=socket_type, in_out='INPUT')

        self.assertSequenceEqual([(s.name, s.item_type) for s in tree.interface.items_tree], [
            ("Output 0", 'SOCKET'),
            ("Input 0", 'SOCKET'),
        ])
        self.assertSequenceEqual([s.name for s in group_node.inputs], [
            "Input 0",
        ])
        self.assertSequenceEqual([s.name for s in group_node.outputs], [
            "Output 0",
        ])
        # XXX currently no panel state access on node instances.
        # self.assertFalse(group_node.panels)

    # Mixed sockets and panels
    def do_test_items_order_mixed_with_panels(self, socket_type):
        tree, group_node = self.make_group_and_instance()

        tree.interface.new_panel("Panel 0")
        tree.interface.new_socket("Input 0", socket_type=socket_type, in_out='INPUT')
        tree.interface.new_socket("Output 0", socket_type=socket_type, in_out='OUTPUT')
        tree.interface.new_panel("Panel 1")
        tree.interface.new_socket("Input 1", socket_type=socket_type, in_out='INPUT')
        tree.interface.new_panel("Panel 2")
        tree.interface.new_socket("Output 1", socket_type=socket_type, in_out='OUTPUT')
        tree.interface.new_panel("Panel 3")

        # Panels after sockets
        self.assertSequenceEqual([(s.name, s.item_type) for s in tree.interface.items_tree], [
            ("Output 0", 'SOCKET'),
            ("Output 1", 'SOCKET'),
            ("Input 0", 'SOCKET'),
            ("Input 1", 'SOCKET'),
            ("Panel 0", 'PANEL'),
            ("Panel 1", 'PANEL'),
            ("Panel 2", 'PANEL'),
            ("Panel 3", 'PANEL'),
        ])
        self.assertSequenceEqual([s.name for s in group_node.inputs], [
            "Input 0",
            "Input 1",
        ])
        self.assertSequenceEqual([s.name for s in group_node.outputs], [
            "Output 0",
            "Output 1",
        ])
        # XXX currently no panel state access on node instances.
        # self.assertSequenceEqual([p.name for p in group_node.panels], [
        #     "Panel 0",
        #     "Panel 1",
        #     "Panel 2",
        #     "Panel 3",
        #     ])

    def do_test_add(self, socket_type):
        tree, group_node = self.make_group_and_instance()

        in0 = tree.interface.new_socket("Input 0", socket_type=socket_type, in_out='INPUT')
        self.assertSequenceEqual(tree.interface.items_tree, [in0])
        self.assertSequenceEqual([s.name for s in group_node.inputs], ["Input 0"])
        self.assertSequenceEqual([s.name for s in group_node.outputs], [])

        out0 = tree.interface.new_socket("Output 0", socket_type=socket_type, in_out='OUTPUT')
        self.assertSequenceEqual(tree.interface.items_tree, [out0, in0])
        self.assertSequenceEqual([s.name for s in group_node.inputs], ["Input 0"])
        self.assertSequenceEqual([s.name for s in group_node.outputs], ["Output 0"])

        panel0 = tree.interface.new_panel("Panel 0")
        self.assertSequenceEqual(tree.interface.items_tree, [out0, in0, panel0])
        self.assertSequenceEqual([s.name for s in group_node.inputs], ["Input 0"])
        self.assertSequenceEqual([s.name for s in group_node.outputs], ["Output 0"])

        # Add items to the panel.
        in1 = tree.interface.new_socket("Input 1", socket_type=socket_type, in_out='INPUT', parent=panel0)
        self.assertSequenceEqual(tree.interface.items_tree, [out0, in0, panel0, in1])
        self.assertSequenceEqual([s.name for s in group_node.inputs], ["Input 0", "Input 1"])
        self.assertSequenceEqual([s.name for s in group_node.outputs], ["Output 0"])

        out1 = tree.interface.new_socket("Output 1", socket_type=socket_type, in_out='OUTPUT', parent=panel0)
        self.assertSequenceEqual(tree.interface.items_tree, [out0, in0, panel0, out1, in1])
        self.assertSequenceEqual([s.name for s in group_node.inputs], ["Input 0", "Input 1"])
        self.assertSequenceEqual([s.name for s in group_node.outputs], ["Output 0", "Output 1"])

    def do_test_remove(self, socket_type):
        tree, group_node = self.make_group_and_instance()

        in0 = tree.interface.new_socket("Input 0", socket_type=socket_type, in_out='INPUT')
        out0 = tree.interface.new_socket("Output 0", socket_type=socket_type, in_out='OUTPUT')
        panel0 = tree.interface.new_panel("Panel 0")
        in1 = tree.interface.new_socket("Input 1", socket_type=socket_type, in_out='INPUT', parent=panel0)
        out1 = tree.interface.new_socket("Output 1", socket_type=socket_type, in_out='OUTPUT', parent=panel0)
        panel1 = tree.interface.new_panel("Panel 1")
        in2 = tree.interface.new_socket("Input 2", socket_type=socket_type, in_out='INPUT', parent=panel1)
        out2 = tree.interface.new_socket("Output 2", socket_type=socket_type, in_out='OUTPUT', parent=panel1)
        panel2 = tree.interface.new_panel("Panel 2")

        self.assertSequenceEqual(tree.interface.items_tree, [out0, in0, panel0, out1, in1, panel1, out2, in2, panel2])
        self.assertSequenceEqual([s.name for s in group_node.inputs], ["Input 0", "Input 1", "Input 2"])
        self.assertSequenceEqual([s.name for s in group_node.outputs], ["Output 0", "Output 1", "Output 2"])

        # Remove from root panel.
        tree.interface.remove(in0)
        self.assertSequenceEqual(tree.interface.items_tree, [out0, panel0, out1, in1, panel1, out2, in2, panel2])
        self.assertSequenceEqual([s.name for s in group_node.inputs], ["Input 1", "Input 2"])
        self.assertSequenceEqual([s.name for s in group_node.outputs], ["Output 0", "Output 1", "Output 2"])

        # Removing a panel should move content to the parent.
        tree.interface.remove(panel0)
        self.assertSequenceEqual(tree.interface.items_tree, [out0, out1, in1, panel1, out2, in2, panel2])
        self.assertSequenceEqual([s.name for s in group_node.inputs], ["Input 1", "Input 2"])
        self.assertSequenceEqual([s.name for s in group_node.outputs], ["Output 0", "Output 1", "Output 2"])

        tree.interface.remove(out0)
        self.assertSequenceEqual(tree.interface.items_tree, [out1, in1, panel1, out2, in2, panel2])
        self.assertSequenceEqual([s.name for s in group_node.inputs], ["Input 1", "Input 2"])
        self.assertSequenceEqual([s.name for s in group_node.outputs], ["Output 1", "Output 2"])

        # Remove content from panel
        tree.interface.remove(out2)
        self.assertSequenceEqual(tree.interface.items_tree, [out1, in1, panel1, in2, panel2])
        self.assertSequenceEqual([s.name for s in group_node.inputs], ["Input 1", "Input 2"])
        self.assertSequenceEqual([s.name for s in group_node.outputs], ["Output 1"])

        # Remove a panel and its content
        tree.interface.remove(panel1, move_content_to_parent=False)
        self.assertSequenceEqual(tree.interface.items_tree, [out1, in1, panel2])
        self.assertSequenceEqual([s.name for s in group_node.inputs], ["Input 1"])
        self.assertSequenceEqual([s.name for s in group_node.outputs], ["Output 1"])

        # Remove empty panel
        tree.interface.remove(panel2)
        self.assertSequenceEqual(tree.interface.items_tree, [out1, in1])
        self.assertSequenceEqual([s.name for s in group_node.inputs], ["Input 1"])
        self.assertSequenceEqual([s.name for s in group_node.outputs], ["Output 1"])

    def do_test_move(self, socket_type):
        tree, group_node = self.make_group_and_instance()

        in0 = tree.interface.new_socket("Input 0", socket_type=socket_type, in_out='INPUT')
        in1 = tree.interface.new_socket("Input 1", socket_type=socket_type, in_out='INPUT', parent=panel0)
        out0 = tree.interface.new_socket("Output 0", socket_type=socket_type, in_out='OUTPUT')
        out1 = tree.interface.new_socket("Output 1", socket_type=socket_type, in_out='OUTPUT', parent=panel0)
        panel0 = tree.interface.new_panel("Panel 0")
        panel1 = tree.interface.new_panel("Panel 1")


class GeometryNodeGroupInterfaceTest(AbstractNodeGroupInterfaceTest, NodeGroupInterfaceTests):
    tree_type = "GeometryNodeTree"
    group_node_type = "GeometryNodeGroup"

    def setUp(self):
        super().setUp()
        self.main_tree = bpy.data.node_groups.new("main", self.tree_type)

    def test_sockets_in_out(self):
        self.do_test_sockets_in_out("NodeSocketFloat")

    def test_all_socket_types(self):
        self.do_test_invalid_socket_type("INVALID_SOCKET_TYPE_11!1")
        self.do_test_socket_type("NodeSocketBool")
        self.do_test_socket_type("NodeSocketCollection")
        self.do_test_socket_type("NodeSocketColor")
        self.do_test_socket_type("NodeSocketFloat")
        self.do_test_socket_type("NodeSocketGeometry")
        self.do_test_socket_type("NodeSocketImage")
        self.do_test_socket_type("NodeSocketInt")
        self.do_test_socket_type("NodeSocketMaterial")
        self.do_test_socket_type("NodeSocketObject")
        self.do_test_socket_type("NodeSocketRotation")
        self.do_test_invalid_socket_type("NodeSocketShader")
        self.do_test_socket_type("NodeSocketString")
        self.do_test_invalid_socket_type("NodeSocketTexture")
        self.do_test_socket_type("NodeSocketVector")
        self.do_test_invalid_socket_type("NodeSocketVirtual")

    def test_items_order_classic(self):
        self.do_test_items_order_classic("NodeSocketFloat")

    def test_items_order_mixed_with_panels(self):
        self.do_test_items_order_mixed_with_panels("NodeSocketFloat")

    def test_add(self):
        self.do_test_add("NodeSocketFloat")

    def test_remove(self):
        self.do_test_remove("NodeSocketFloat")


class ShaderNodeGroupInterfaceTest(AbstractNodeGroupInterfaceTest, NodeGroupInterfaceTests):
    tree_type = "ShaderNodeTree"
    group_node_type = "ShaderNodeGroup"

    def setUp(self):
        super().setUp()
        self.material = bpy.data.materials.new("test")
        self.main_tree = self.material.node_tree

    def test_invalid_socket_type(self):
        self.do_test_invalid_socket_type("INVALID_SOCKET_TYPE_11!1")

    def test_sockets_in_out(self):
        self.do_test_sockets_in_out("NodeSocketFloat")

    def test_all_socket_types(self):
        self.do_test_socket_type("NodeSocketBool")
        self.do_test_invalid_socket_type("NodeSocketCollection")
        self.do_test_socket_type("NodeSocketColor")
        self.do_test_socket_type("NodeSocketFloat")
        self.do_test_invalid_socket_type("NodeSocketGeometry")
        self.do_test_invalid_socket_type("NodeSocketImage")
        self.do_test_socket_type("NodeSocketInt")
        self.do_test_invalid_socket_type("NodeSocketMaterial")
        self.do_test_invalid_socket_type("NodeSocketObject")
        self.do_test_invalid_socket_type("NodeSocketRotation")
        self.do_test_socket_type("NodeSocketShader")
        self.do_test_invalid_socket_type("NodeSocketString")
        self.do_test_invalid_socket_type("NodeSocketTexture")
        self.do_test_socket_type("NodeSocketVector")
        self.do_test_invalid_socket_type("NodeSocketVirtual")

    def test_items_order_classic(self):
        self.do_test_items_order_classic("NodeSocketFloat")

    def test_items_order_mixed_with_panels(self):
        self.do_test_items_order_mixed_with_panels("NodeSocketFloat")

    def test_add(self):
        self.do_test_add("NodeSocketFloat")

    def test_remove(self):
        self.do_test_remove("NodeSocketFloat")


class CompositorNodeGroupInterfaceTest(AbstractNodeGroupInterfaceTest, NodeGroupInterfaceTests):
    tree_type = "CompositorNodeTree"
    group_node_type = "CompositorNodeGroup"

    def setUp(self):
        super().setUp()
        self.scene = bpy.data.scenes.new("test")
        self.main_tree = bpy.data.node_groups.new("test node tree", "CompositorNodeTree")
        self.scene.compositing_node_group = self.main_tree

    def test_invalid_socket_type(self):
        self.do_test_invalid_socket_type("INVALID_SOCKET_TYPE_11!1")

    def test_sockets_in_out(self):
        self.do_test_sockets_in_out("NodeSocketFloat")

    def test_all_socket_types(self):
        self.do_test_socket_type("NodeSocketBool")
        self.do_test_invalid_socket_type("NodeSocketCollection")
        self.do_test_socket_type("NodeSocketColor")
        self.do_test_socket_type("NodeSocketFloat")
        self.do_test_invalid_socket_type("NodeSocketGeometry")
        self.do_test_invalid_socket_type("NodeSocketImage")
        self.do_test_socket_type("NodeSocketInt")
        self.do_test_invalid_socket_type("NodeSocketMaterial")
        self.do_test_invalid_socket_type("NodeSocketObject")
        self.do_test_invalid_socket_type("NodeSocketRotation")
        self.do_test_invalid_socket_type("NodeSocketShader")
        self.do_test_socket_type("NodeSocketString")
        self.do_test_invalid_socket_type("NodeSocketTexture")
        self.do_test_socket_type("NodeSocketVector")
        self.do_test_invalid_socket_type("NodeSocketVirtual")

    def test_items_order_classic(self):
        self.do_test_items_order_classic("NodeSocketFloat")

    def test_items_order_mixed_with_panels(self):
        self.do_test_items_order_mixed_with_panels("NodeSocketFloat")

    def test_add(self):
        self.do_test_add("NodeSocketFloat")

    def test_remove(self):
        self.do_test_remove("NodeSocketFloat")


class NodeTreeItemsIteratorTest(AbstractNodeGroupInterfaceTest, NodeGroupInterfaceTests):
    tree_type = "ShaderNodeTree"
    group_node_type = "ShaderNodeGroup"

    def setUp(self):
        super().setUp()
        self.material = bpy.data.materials.new("test")
        self.main_tree = self.material.node_tree

    # Regression test for changes while iterating over tree interface items (#143551).
    # The iterator should remain valid when changing properties of a tree item.
    def test_items_iterator(self):
        tree, group_node = self.make_group_and_instance()

        tree.interface.new_socket("Input 0", socket_type="NodeSocketFloat", in_out='INPUT')
        tree.interface.new_socket("Input 1", socket_type="NodeSocketBool", in_out='INPUT')
        # The cache vector has a fixed buffer for small sizes, add enough sockets to force reallocation.
        for i in range(20):
            tree.interface.new_socket(f"Input {2+i}", socket_type="NodeSocketColor", in_out='INPUT')

        # Iterate over items and change properties. The loop iterator must remain valid.
        for item in tree.interface.items_tree:
            if item.socket_type == "NodeSocketFloat":
                item.default_value = 500.0
            elif item.socket_type == "NodeSocketColor":
                item.default_value = (1, 0, 0, 1)
            elif item.socket_type == "NodeSocketBool":
                item.default_value = True


def main():
    global args
    import argparse

    if '--' in sys.argv:
        argv = [sys.argv[0]] + sys.argv[sys.argv.index('--') + 1:]
    else:
        argv = sys.argv

    parser = argparse.ArgumentParser()
    parser.add_argument('--testdir', required=True, type=pathlib.Path)
    args, remaining = parser.parse_known_args(argv)

    unittest.main(argv=remaining)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2025 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

# ./blender.bin --background --factory-startup --python tests/python/bl_pyapi_bpy_app_tempdir.py -- --verbose

# NOTE(ideasman42):
#
# - Creating a directory without write permissions across all supported platforms
#   is it not trivial, for the purpose of these tests simply create a file as a way of
#   pointing to a temporary path which can't have sub-directories created under it.
#
# - Internally, when no temporary path is set the hard coded path `/tmp/` is used.
#   For the purpose of the tests being complete - control over this path would be needed too.
#
# - When changing the temp directory & on exit
#   Blender's *session* temp directory is recursively removed.
#
#   These tests were added as part of a fix for a serious flaw (see #144042)
#   which would recursively delete the users `C:\` since the tests aren't sand-boxed
#   all of the following tests reference paths within a newly creating temporary directory
#   instead of referencing root paths to avoid risks for developers who run tests.
#
# - A related test exists `BLI_tempfile_test.cc` however this doesn't deal with
#   Blender's user preferences and fallbacks used in Blender.
#

__all__ = (
    "main",
)

import os
import unittest
import tempfile


is_win32 = os.name == "nt"
TEMP_ENV = "TEMP" if is_win32 else "TMPDIR"


def system_temp_set(path: str) -> None:
    os.environ[TEMP_ENV] = path


def prefs_temp_set(path: str) -> None:
    import bpy  # type: ignore
    bpy.context.preferences.filepaths.temporary_directory = path


def prefs_temp_get() -> str:
    import bpy
    result = bpy.context.preferences.filepaths.temporary_directory
    assert isinstance(result, str)
    return result


def blender_tempdir_session_get() -> str:
    import bpy
    result = bpy.app.tempdir
    assert isinstance(result, str)
    return result


def empty_file(path: str) -> None:
    with open(path, 'wb') as _fh:
        pass


def commonpath_safe(paths: list[str]) -> str:
    if is_win32:
        try:
            return os.path.commonpath(paths)
        except ValueError:
            # Workaround error on Windows.
            # ValueError: Paths don't have the same drive
            return ""

    return os.path.commonpath(paths)


class TestTempDir(unittest.TestCase):

    def setUp(self) -> None:
        print(prefs_temp_get())
        assert prefs_temp_get() == ""
        system_temp_set("")
        prefs_temp_set("")

    def tearDown(self) -> None:
        prefs_temp_set("")
        system_temp_set("")

    def test_fallback(self) -> None:
        # Set a file for preferences & system temp, ensure neither are used.
        with tempfile.TemporaryDirectory() as tempdir:

            temp_sys = os.path.join(tempdir, "a_sys")
            temp_bpy = os.path.join(tempdir, "b_bpy")

            empty_file(temp_sys)
            empty_file(temp_bpy)

            system_temp_set(temp_sys)
            prefs_temp_set(temp_bpy)

            temp_session = blender_tempdir_session_get()

            # Ensure neither are used.
            # This will try to use `/tmp/`.
            commonpath_test = commonpath_safe([temp_sys, temp_session])
            self.assertFalse(commonpath_test and os.path.samefile(temp_sys, commonpath_test))
            commonpath_test = commonpath_safe([temp_bpy, temp_session])
            self.assertFalse(commonpath_test and os.path.samefile(temp_bpy, commonpath_test))

    def test_system(self) -> None:
        # Set an file as the preferences temp directory, ensure the system path is used.
        with tempfile.TemporaryDirectory() as tempdir:

            temp_sys = os.path.join(tempdir, "a_sys")
            temp_bpy = os.path.join(tempdir, "b_bpy")

            os.mkdir(temp_sys)
            empty_file(temp_bpy)

            system_temp_set(temp_sys)
            prefs_temp_set(temp_bpy)

            temp_session = blender_tempdir_session_get()

            self.assertTrue(os.path.samefile(temp_sys, os.path.commonpath([temp_sys, temp_session])))

    def test_prefs(self) -> None:
        # Set an file as the system temp directory, ensure the preferences path is used.
        with tempfile.TemporaryDirectory() as tempdir:

            temp_sys = os.path.join(tempdir, "a_sys")
            temp_bpy = os.path.join(tempdir, "b_bpy")

            empty_file(temp_sys)
            os.mkdir(temp_bpy)

            system_temp_set(temp_sys)
            prefs_temp_set(temp_bpy)

            temp_session = blender_tempdir_session_get()

            self.assertTrue(os.path.samefile(temp_bpy, os.path.commonpath([temp_bpy, temp_session])))

    def test_system_to_prefs(self) -> None:
        with tempfile.TemporaryDirectory() as tempdir:
            temp_sys = os.path.join(tempdir, "a_sys")
            temp_bpy = os.path.join(tempdir, "b_bpy")

            os.mkdir(temp_sys)
            empty_file(temp_bpy)

            system_temp_set(temp_sys)
            prefs_temp_set(temp_bpy)

            temp_session = blender_tempdir_session_get()
            self.assertTrue(os.path.samefile(temp_sys, os.path.commonpath([temp_sys, temp_session])))

            # Now set the preferences and ensure the previous directory gets purged and the new one set.
            os.unlink(temp_bpy)
            os.mkdir(temp_bpy)

            # Ensure the preferences path is now used.
            prefs_temp_set(temp_bpy)

            self.assertFalse(os.path.exists(temp_session))
            temp_session = blender_tempdir_session_get()
            self.assertTrue(os.path.samefile(temp_bpy, os.path.commonpath([temp_bpy, temp_session])))

    def test_prefs_to_system(self) -> None:
        with tempfile.TemporaryDirectory() as tempdir:
            temp_sys = os.path.join(tempdir, "a_sys")
            temp_bpy = os.path.join(tempdir, "b_bpy")

            empty_file(temp_sys)
            os.mkdir(temp_bpy)

            system_temp_set(temp_sys)
            prefs_temp_set(temp_bpy)

            temp_session = blender_tempdir_session_get()
            self.assertTrue(os.path.samefile(temp_bpy, os.path.commonpath([temp_bpy, temp_session])))

            # Now set the preferences and ensure the previous directory gets purged and the new one set.
            os.unlink(temp_sys)
            os.mkdir(temp_sys)

            # Ensure the system path is now used.
            prefs_temp_set(temp_sys)

            self.assertFalse(os.path.exists(temp_session))
            temp_session = blender_tempdir_session_get()
            self.assertTrue(os.path.samefile(temp_sys, os.path.commonpath([temp_sys, temp_session])))


def main():
    import sys
    unittest.main(argv=[__file__] + (sys.argv[sys.argv.index("--") + 1:] if "--" in sys.argv else []))


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2024 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

from .gdb import Value


class PrettyPrinter:
    def __init__(self, name: str, subprinters=None):
        pass

    def __call__(self, value: Value):
        pass



bl_info = {
    "name": "Example Add-on Preferences",
    "author": "Your Name Here",
    "version": (1, 0),
    "blender": (2, 65, 0),
    "location": "SpaceBar Search -> Add-on Preferences Example",
    "description": "Example Add-on",
    "warning": "",
    "doc_url": "",
    "tracker_url": "",
    "category": "Object",
}


import bpy
from bpy.types import Operator, AddonPreferences
from bpy.props import StringProperty, IntProperty, BoolProperty


class ExampleAddonPreferences(AddonPreferences):
    # This must match the add-on name, use `__package__`
    # when defining this for add-on extensions or a sub-module of a Python package.
    bl_idname = __name__

    filepath: StringProperty(
        name="Example File Path",
        subtype='FILE_PATH',
    )
    number: IntProperty(
        name="Example Number",
        default=4,
    )
    boolean: BoolProperty(
        name="Example Boolean",
        default=False,
    )

    def draw(self, context):
        layout = self.layout
        layout.label(text="This is a preferences view for our add-on")
        layout.prop(self, "filepath")
        layout.prop(self, "number")
        layout.prop(self, "boolean")


class OBJECT_OT_addon_prefs_example(Operator):
    """Display example preferences"""
    bl_idname = "object.addon_prefs_example"
    bl_label = "Add-on Preferences Example"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        preferences = context.preferences
        addon_prefs = preferences.addons[__name__].preferences

        info = "Path: {:s}, Number: {:d}, Boolean {!r}".format(
            addon_prefs.filepath, addon_prefs.number, addon_prefs.boolean,
        )
        self.report({'INFO'}, info)
        print(info)

        return {'FINISHED'}


# Registration
def register():
    bpy.utils.register_class(OBJECT_OT_addon_prefs_example)
    bpy.utils.register_class(ExampleAddonPreferences)


def unregister():
    bpy.utils.unregister_class(OBJECT_OT_addon_prefs_example)
    bpy.utils.unregister_class(ExampleAddonPreferences)



"""
Drawing Text to an Image
++++++++++++++++++++++++

Example showing how text can be draw into an image.
This can be done by binding an image buffer (:mod:`imbuf`) to the font's ID.
"""

import blf
import imbuf

image_size = 512, 512
font_size = 20

ibuf = imbuf.new(image_size)

font_id = blf.load("/path/to/font.ttf")

blf.color(font_id, 1.0, 1.0, 1.0, 1.0)
blf.size(font_id, font_size)
blf.position(font_id, 0, image_size[0] - font_size, 0)

blf.enable(font_id, blf.WORD_WRAP)
blf.word_wrap(font_id, image_size[0])

with blf.bind_imbuf(font_id, ibuf, display_name="sRGB"):
    blf.draw_buffer(font_id, "Lots of wrapped text. " * 50)

imbuf.write(ibuf, filepath="/path/to/image.png")



# SPDX-FileCopyrightText: 2025 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later */

__all__ = (
    "main",
)

import unittest
import sys
import pathlib
import numpy as np

import bpy

"""
blender -b --factory-startup --python tests/python/bl_sculpt.py -- --testdir tests/files/sculpting/
"""

args = None


def set_view3d_context_override(context_override):
    """
    Set context override to become the first viewport in the active workspace

    The ``context_override`` is expected to be a copy of an actual current context
    obtained by `context.copy()`
    """

    for area in context_override["screen"].areas:
        if area.type != 'VIEW_3D':
            continue
        for space in area.spaces:
            if space.type != 'VIEW_3D':
                continue
            for region in area.regions:
                if region.type != 'WINDOW':
                    continue
                context_override["area"] = area
                context_override["region"] = region


class MaskByColorTest(unittest.TestCase):
    def setUp(self):
        bpy.ops.wm.open_mainfile(filepath=str(args.testdir / "plane_with_red_circle.blend"), load_ui=False)

        self.context_override = bpy.context.copy()
        set_view3d_context_override(self.context_override)
        bpy.ops.ed.undo_push()

    def test_off_grid_returns_cancelled(self):
        """Test that operator does not run when the cursor is not on the mesh."""

        with bpy.context.temp_override(**self.context_override):
            location = (0, 0)
            ret_val = bpy.ops.sculpt.mask_by_color(location=location)

            self.assertEqual({'CANCELLED'}, ret_val)

        mesh = bpy.context.object.data
        self.assertFalse('.sculpt_mask' in mesh.attributes.keys(), "Mesh should not have the .sculpt_mask attribute!")

    def test_on_circle_masks_red_vertices(self):
        """Test that the operator only masks red vertices on the mesh."""

        with bpy.context.temp_override(**self.context_override):
            location = (int(self.context_override['area'].width / 2), int(self.context_override['area'].height / 2))
            ret_val = bpy.ops.sculpt.mask_by_color(location=location)

            self.assertEqual({'FINISHED'}, ret_val)

        mesh = bpy.context.object.data
        color_attr = mesh.attributes['Color']
        mask_attr = mesh.attributes['.sculpt_mask']

        num_vertices = mesh.attributes.domain_size('POINT')

        color_data = np.zeros((num_vertices, 4), dtype=np.float32)
        color_attr.data.foreach_get('color', np.ravel(color_data))

        mask_data = np.zeros(num_vertices, dtype=np.float32)
        mask_attr.data.foreach_get('value', mask_data)

        for i in range(num_vertices):
            # If either of the green or blue components are less than 1 (i.e. the vertex is the red part of the image instead of
            # the white background), then that vertex should also be masked.
            if color_data[i][1] < 0.4 and color_data[i][2] < 0.4:
                self.assertTrue(mask_data[i] > 0.0, f"Vertex {i} should be masked ({color_data[i]}) -> {mask_data[i]}")
            else:
                self.assertTrue(mask_data[i] < 0.1,
                                f"Vertex {i} should not be masked ({color_data[i]}) -> {mask_data[i]}")


class MaskFromCavityTest(unittest.TestCase):
    def setUp(self):
        bpy.ops.wm.open_mainfile(filepath=str(args.testdir / "plane_with_valley.blend"), load_ui=False)
        bpy.ops.ed.undo_push()

    def test_operator_masks_low_vertices(self):
        """Test that the operator applies a full mask value to any elements that are part of the cavity."""

        ret_val = bpy.ops.sculpt.mask_from_cavity()

        self.assertEqual({'FINISHED'}, ret_val)

        mesh = bpy.context.object.data
        position_attr = mesh.attributes['position']
        mask_attr = mesh.attributes['.sculpt_mask']

        num_vertices = mesh.attributes.domain_size('POINT')

        position_data = np.zeros((num_vertices, 3), dtype=np.float32)
        position_attr.data.foreach_get('vector', np.ravel(position_data))

        mask_data = np.zeros(num_vertices, dtype=np.float32)
        mask_attr.data.foreach_get('value', mask_data)

        for i in range(num_vertices):
            if position_data[i][2] < 0.0:
                self.assertEqual(
                    mask_data[i],
                    1.0,
                    f"Vertex {i} should be fully masked ({position_data[i]}) -> {mask_data[i]}")
            else:
                self.assertNotEqual(mask_data[i], 1.0,
                                    f"Vertex {i} should not be fully masked ({position_data[i]}) -> {mask_data[i]}")


class DetailFloodFillTest(unittest.TestCase):
    def setUp(self):
        bpy.ops.wm.read_factory_settings(use_empty=True)
        bpy.ops.ed.undo_push()

        bpy.ops.mesh.primitive_cube_add()
        bpy.ops.sculpt.sculptmode_toggle()
        bpy.ops.sculpt.dynamic_topology_toggle()

    def test_operator_subdivides_mesh(self):
        """Test that the operator generates a mesh with appropriately sized edges."""

        max_edge_length = 1.0
        # Based on the detail_size::EDGE_LENGTH_MIN_FACTOR constant
        min_edge_length = max_edge_length * 0.4

        bpy.context.scene.tool_settings.sculpt.detail_type_method = 'CONSTANT'
        bpy.context.scene.tool_settings.sculpt.constant_detail_resolution = max_edge_length

        ret_val = bpy.ops.sculpt.detail_flood_fill()
        self.assertEqual({'FINISHED'}, ret_val)

        # Toggle to ensure the mesh data is refreshed.
        bpy.ops.sculpt.dynamic_topology_toggle()

        mesh = bpy.context.object.data
        for edge in mesh.edges:
            v0 = mesh.vertices[edge.vertices[0]]
            v1 = mesh.vertices[edge.vertices[1]]

            length = (v0.co - v1.co).length

            self.assertGreaterEqual(
                length,
                min_edge_length,
                f"Edge between {v0.index} and {v1.index} should be longer than minimum length")
            self.assertLessEqual(
                length,
                max_edge_length,
                f"Edge between {v0.index} and {v1.index} should be shorter than maximum length")


def main():
    global args
    import argparse

    argv = [sys.argv[0]]
    if '--' in sys.argv:
        argv += sys.argv[sys.argv.index('--') + 1:]

    parser = argparse.ArgumentParser()
    parser.add_argument('--testdir', required=True, type=pathlib.Path)

    args, remaining = parser.parse_known_args(argv)

    unittest.main(argv=remaining)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2023 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""
This test calls into the path pattern matching logic of ``blender_ext`` directly.
"""
import unittest
import os

from typing import (
    Any,
)
from collections.abc import (
    Sequence,
)


CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
BASE_DIR = os.path.normpath(os.path.join(CURRENT_DIR, ".."))


# Don't import as module, instead load the class.
def execfile(filepath: str, *, name: str = "__main__") -> dict[str, Any]:
    global_namespace = {"__file__": filepath, "__name__": name}
    with open(filepath, encoding="utf-8") as fh:
        # pylint: disable-next=exec-used
        exec(compile(fh.read(), filepath, 'exec'), global_namespace)
    return global_namespace


PathPatternMatch = execfile(os.path.join(BASE_DIR, "cli", "blender_ext.py"), name="blender_ext")["PathPatternMatch"]
assert isinstance(PathPatternMatch, type)


class TestPathMatch_MixIn:

    def match_paths(
            self,
            expected_paths: list[tuple[bool, str]],
            path_pattern: Sequence[str] | PathPatternMatch,  # type: ignore
    ) -> list[tuple[bool, str]]:
        result = []
        if not isinstance(path_pattern, PathPatternMatch):
            path_pattern = PathPatternMatch(path_pattern)
        assert hasattr(path_pattern, "test_path")
        for _success_expected, path in expected_paths:
            success_actual = path_pattern.test_path(path)
            # Un-comment to pin-point the exact error.
            # `self.assertEqual(success_actual, success_expected)`
            result.append((success_actual, path))
        return result

    def match_paths_for_cmp(
            self,
            expected_paths: list[tuple[bool, str]],
            path_pattern: Sequence[str] | PathPatternMatch,  # type: ignore
    ) -> tuple[
        list[tuple[bool, str]],
        list[tuple[bool, str]],
    ]:
        return self.match_paths(expected_paths, path_pattern), expected_paths


class TestSingle(unittest.TestCase, TestPathMatch_MixIn):

    def test_directory(self) -> None:
        pattern = PathPatternMatch(["__pycache__/"])
        self.assertEqual(*self.match_paths_for_cmp([(True, "__pycache__/")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(True, "a/__pycache__/")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(True, "a/__pycache__/any")], pattern))

        # Not a directory.
        self.assertEqual(*self.match_paths_for_cmp([(False, "__pycache__")], pattern))

        # Not exact matches.
        self.assertEqual(*self.match_paths_for_cmp([(False, "__pycache_/")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(False, "__pycache___/")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(False, "___pycache__/")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(False, "_pycache__/")], pattern))

        # Not exact matches (sub-directory).
        self.assertEqual(*self.match_paths_for_cmp([(False, "a/__pycache_/")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(False, "a/__pycache___/")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(False, "a/___pycache__/")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(False, "a/_pycache__/")], pattern))

    def test_file_extension(self) -> None:
        pattern = PathPatternMatch(["*.pyc"])
        self.assertEqual(*self.match_paths_for_cmp([(True, ".pyc")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(True, "a.pyc")], pattern))

        self.assertEqual(*self.match_paths_for_cmp([(True, "a/.pyc")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(True, "a/a.pyc")], pattern))

        # Not exact matches.
        self.assertEqual(*self.match_paths_for_cmp([(False, ".pycx")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(False, "a.pycx")], pattern))

        self.assertEqual(*self.match_paths_for_cmp([(False, ".pyc/x")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(False, "a.pyc/x")], pattern))

        # Not exact matches (sub-directory).
        self.assertEqual(*self.match_paths_for_cmp([(False, "a/.pycx")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(False, "a/a.pycx")], pattern))

        self.assertEqual(*self.match_paths_for_cmp([(False, "a/.pyc/x")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(False, "a/a.pyc/x")], pattern))

    def test_double_star(self) -> None:
        pattern = PathPatternMatch(["a/**/b"])
        self.assertEqual(*self.match_paths_for_cmp([(True, "a/x/b")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(True, "a/x/y/z/b")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(True, "a/b")], pattern))

        self.assertEqual(*self.match_paths_for_cmp([(False, "a_/x/b")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(False, "a_/x/y/z/b")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(False, "a_/b")], pattern))

        self.assertEqual(*self.match_paths_for_cmp([(False, "a/x/_b")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(False, "a/x/y/z/_b")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(False, "a/_b")], pattern))

    def test_double_star_directory(self) -> None:
        pattern = PathPatternMatch(["a/**/b/"])
        self.assertEqual(*self.match_paths_for_cmp([(True, "a/x/b/")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(True, "a/x/y/z/b/")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(True, "a/b/")], pattern))

    def test_root_directory(self) -> None:
        pattern = PathPatternMatch(["/build"])
        self.assertEqual(*self.match_paths_for_cmp([(True, "build/")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(True, "build")], pattern))

        # Not absolute.
        self.assertEqual(*self.match_paths_for_cmp([(False, "a/build/")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(False, "a/build")], pattern))

        self.assertEqual(*self.match_paths_for_cmp([(False, "build_/")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(False, "build_")], pattern))

    def test_root_directory_file(self) -> None:
        pattern = PathPatternMatch(["/a/build.txt"])
        self.assertEqual(*self.match_paths_for_cmp([(True, "a/build.txt")], pattern))

        # Not root relative.
        self.assertEqual(*self.match_paths_for_cmp([(False, "b/a/build.txt")], pattern))

    def test_root_directory_file_star(self) -> None:
        pattern = PathPatternMatch(["/test/*.txt"])
        self.assertEqual(*self.match_paths_for_cmp([(True, "test/a.txt")], pattern))

        self.assertEqual(*self.match_paths_for_cmp([(False, "test/b/c.txt")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(False, "b/test/c.txt")], pattern))


class TestMulti(unittest.TestCase, TestPathMatch_MixIn):

    def test_negate(self) -> None:
        pattern = PathPatternMatch([
            "data.*",
            "!data.csv",
        ])
        self.assertEqual(*self.match_paths_for_cmp([(True, "test/data.txt")], pattern))

        self.assertEqual(*self.match_paths_for_cmp([(False, "test/data.csv")], pattern))

    def test_negate_spesific(self) -> None:
        pattern = PathPatternMatch([
            "data.*",
            "!/test/data.csv",
        ])

        self.assertEqual(*self.match_paths_for_cmp([(True, "other/data.csv")], pattern))
        self.assertEqual(*self.match_paths_for_cmp([(True, "other/data.txt")], pattern))

        self.assertEqual(*self.match_paths_for_cmp([(False, "test/data.csv")], pattern))


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: 2017-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

# ############################################################
# Importing - Same For All Render Layer Tests
# ############################################################

import unittest

from view_layer_common import (
    Clay,
    ViewLayerTesting,
    setup_extra_arguments,
)


# ############################################################
# Testing
# ############################################################

class UnitTesting(ViewLayerTesting):
    @unittest.skip("Uses the clay engine, that is removed #55454")
    def test_render_settings(self):
        """
        See if the depsgraph evaluation is correct
        """
        clay = Clay(extra_kid_layer=True)
        self.assertEqual(clay.get('object', 'matcap_icon'), '01')

        clay.set('kid', 'matcap_icon', '02')
        clay.set('extra', 'matcap_icon', '04')
        self.assertEqual(clay.get('object', 'matcap_icon'), '04')


# ############################################################
# Main - Same For All Render Layer Tests
# ############################################################

if __name__ == '__main__':
    UnitTesting._extra_arguments = setup_extra_arguments(__file__)
    unittest.main()



# SPDX-FileCopyrightText: 2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

'''
Usage:
    python discover_nodes.py
        <sources/root>
        <path/to/output.cc>
        <generated_function_name>
        <source>...

The goal is to make it easy for nodes to register themselves without having to have
a central place that registers all nodes manually. A node can use this mechanism by
invoking `NOD_REGISTER_NODE(register_function_name)`.

This scripts finds all those macro invocations generates code that calls the functions.
'''
__all__ = (
    "main",
)

import os
import re
import sys


def filepath_is_older(filepath_test: str, filepath_compare: tuple[str, ...]) -> bool:
    import stat
    mtime = os.stat(filepath_test)[stat.ST_MTIME]
    for filepath_other in filepath_compare:
        if mtime < os.stat(filepath_other)[stat.ST_MTIME]:
            return True
    return False


def main() -> int:
    # The build system requires the generated file to be touched if any files used to generate it are newer.
    try:
        sys.argv.remove("--use-makefile-workaround")
        use_makefile_workaround = True
    except ValueError:
        use_makefile_workaround = False

    # NOTE: avoid `pathlib`, pulls in many modules indirectly, path handling is simple enough.
    source_root = sys.argv[1]
    output_cc_file = sys.argv[2]
    function_to_generate = sys.argv[3]
    source_cc_files = [
        os.path.join(source_root, path)
        for path in sys.argv[4:]
        if path.endswith(".cc")
    ]

    macro_name = "NOD_REGISTER_NODE"
    discover_suffix = "_discover"

    include_lines: list[str] = []
    decl_lines: list[str] = []
    func_lines: list[str] = []

    # Add forward declaration to avoid warning.
    func_lines.append(f"void {function_to_generate}();")
    func_lines.append(f"void {function_to_generate}()")
    func_lines.append("{")

    # Use a single regular expression to search for opening name-spaces, closing name-spaces
    # and macro invocations. This makes it easy to iterate over the matches in order.
    re_namespace_begin = r"^namespace ([\w:]+) \{"
    re_namespace_end = r"^\}  // namespace ([\w:]+)"
    re_macro = r"MACRO\((\w+)\)".replace("MACRO", macro_name)
    re_all = f"({re_namespace_begin})|({re_namespace_end})|({re_macro})"
    re_all_compiled = re.compile(re_all, flags=re.MULTILINE)

    for path in source_cc_files:
        # Read the source code.
        with open(path, "r", encoding="utf-8") as fh:
            code = fh.read()

        # Keeps track of the current name-space we're in.
        namespace_parts: list[str] = []

        for match in re_all_compiled.finditer(code):
            if entered_namespace := match.group(2):
                # Enter a (nested) name-space.
                namespace_parts += entered_namespace.split("::")
            elif exited_namespace := match.group(4):
                # Exit a (nested) name-space.
                del namespace_parts[-len(exited_namespace.split("::")):]
            elif function_name := match.group(6):
                # Macro invocation in the current name-space.
                namespace_str = "::".join(namespace_parts)
                # Add suffix so that this refers to the function created by the macro.
                auto_run_name = function_name + discover_suffix

                # Declare either outside of any named name-space or in a (nested) name-space.
                # Can't declare it in an anonymous name-space because that would make the
                # declared function static.
                if namespace_str:
                    decl_lines.append(f"namespace {namespace_str} {{")
                decl_lines.append(f"void {auto_run_name}();")
                if namespace_str:
                    decl_lines.append("}")

                # Call the function.
                func_lines.append(f"  {namespace_str}::{auto_run_name}();")

    func_lines.append("}")

    # Write the generated code if it changed. If the newly generated code is the same as before,
    # don't overwrite the existing file to avoid unnecessary rebuilds.
    try:
        with open(output_cc_file, "r", encoding="utf-8") as fh:
            old_generated_code = fh.read()
    except Exception:
        old_generated_code = ""
    new_generated_code = "\n".join(include_lines + decl_lines + [""] + func_lines)

    if old_generated_code != new_generated_code:
        with open(output_cc_file, "w", encoding="utf-8") as fh:
            fh.write(new_generated_code)
    elif use_makefile_workaround and filepath_is_older(output_cc_file, (__file__, *source_cc_files)):
        # If the generated file is older than this command, this file would be generated every time.
        os.utime(output_cc_file)

    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# SPDX-FileCopyrightText: 2025 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""Self-bootstrapping script to run the OpenAPI-to-dataclasses code generator.

Run this via `<your buildtool> generate_datamodels` in your build directory.

This script creates its own virtualenv, installs its dependencies, and then runs
the code generator. It processes OpenAPI spec files in YAML format (see
`YAML_PATHS` below) to generate Python source files. Each `xxx.yaml` file will
produce an `xxx.py` file in the same directory. These Python files also include
the OpenAPI spec, as a Python dict.

The generated Python files are tracked by Git. This generator is NOT part of the
regular Blender build process, and only needs to be run when any of the YAML
files change.
"""

__all__ = (
    "main",
)

import argparse
from pathlib import Path
import sys
import time

# Paths of the OpenAPI YAML files to convert to Python code. These are relative
# to Blender's top level source directory.
#
# The generated Python files will be written to the same path, just with the
# `.py` suffix.
#
# When adding a file here, make sure it is named `..._openapi.yaml`. That way
# the corresponding `.py`` file is automatically marked as 'generated' in
# `.gitattributes`.
YAML_PATHS = [
    "scripts/modules/_bpy_internal/assets/remote_library_listing/blender_asset_library_openapi.yaml",
]

# Packages to install in the virtualenv. These are only necessary to run this
# generator. The generated code does not depend on these.
REQUIREMENTS = [
    "datamodel-code-generator ~= 0.28.2",
    "PyYAML ~= 6.0.2",
]

# These arguments are quite likely to be used for all code generated with this
# generator, also later when we use this approach in other areas.
COMMON_ARGS = [
    # Because of the Blender code standard:
    "--use-double-quotes",

    # Make it strict unless there's a good reason not to:
    "--strict-nullable",

    # Ignore unknown fields in the parsed JSON. This way, the code generated now
    # has a chance of being be compatible with future versions of the schema (at
    # least, when that future version just adds new stuff).
    "--allow-extra-fields",

    # Automatically target the currently-running version of Python:
    f"--target-python-version={sys.version_info.major}.{sys.version_info.minor}",

    # Use `list[T]` instead of `typing.List[T]`:
    "--use-standard-collections",

    # Because we use dataclasses:
    "--output-model-type", "dataclasses.dataclass",

    # Work around https://github.com/koxudaxi/datamodel-code-generator/issues/1870#issuecomment-2775689249
    "--use-annotated",

    # Remove the "generated on" timestamp from the output, so that running the
    # generator is idempotent.
    "--disable-timestamp",
]


CUSTOM_FILE_HEADER = """
# SPDX-FileCopyrightText: {year!s} Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# Generated by datamodel-codegen:
#   source filename:  {source_path.name!s}
"""


def main() -> None:
    """Run the datamodel code generator."""

    # Late import, as this is only available once inside the virtualenv.
    import yaml

    argparser = argparse.ArgumentParser(description="Run the datamodel code generator.")
    argparser.add_argument('source_root', type=Path, help="The root of Blender's source directory")
    args = argparser.parse_args(sys.argv[1:])
    root_path: Path = args.source_root.resolve()

    if not root_path.is_dir():
        raise SystemExit("Path {!s} should be a directory".format(root_path))

    print("Generating data model files:")

    py_paths: list[Path] = []
    for yaml_relpath in YAML_PATHS:
        yaml_path = root_path / yaml_relpath
        py_path = yaml_path.with_suffix(".py")
        py_paths.append(py_path)

        print(f"  {yaml_path.relative_to(root_path)} -> {py_path.name}")

        _generate_datamodel(
            in_path=yaml_path,
            in_type="openapi",
            out_path=py_path,
        )

        # Append the OpenAPI specification as Python code. This is necessary to
        # reference for runtime validation. Having it available as Python
        # dictionary is easier than having to parse it from YAML or JSON later.
        with yaml_path.open() as yamlfile:
            openapi_spec = yaml.safe_load(yamlfile)
        with py_path.open("a") as outfile:
            print(file=outfile)
            print("# This OpenAPI specification was used to generate the above code.", file=outfile)
            print("# It is here so that Blender does not have to parse the YAML file.", file=outfile)
            print("OPENAPI_SPEC = {!r}".format(openapi_spec), file=outfile)

    # Make sure that output from subprocesses is flushed, before outputting more
    # below. This prevents stderr and stdout going out of sync, ensuring things
    # are shown in chronological order (i.e. generating files before
    # reformatting them).
    sys.stderr.flush()
    sys.stdout.flush()

    # Format the generated Python code. Autopep8 (used by Blender) does not
    # seem to re-wrap long lines, so that's why this script relies on running
    # ruff first.
    print("Formatting Python files")
    py_paths_as_str = [str(path) for path in py_paths]
    subprocess.run(
        ["make", "format", "PATHS={}".format(" ".join(py_paths_as_str))],
        cwd=root_path,
        check=True,
    )

    print("Done generating data model files!")


def _generate_datamodel(in_path: Path, in_type: str, out_path: Path) -> None:
    """Run datamodel-codegen."""

    # `type: ignore` to ignore warnings that this module cannot be imported. Python checkers
    # won't understand this is run from a self-managed virtualenv.
    from datamodel_code_generator.__main__ import main as codegen_main  # type: ignore
    from datamodel_code_generator.__main__ import Exit  # type: ignore

    header = CUSTOM_FILE_HEADER.strip().format(
        year=time.localtime().tm_year,
        source_path=in_path,
    )

    args = [
        *COMMON_ARGS,
        "--input", str(in_path),
        "--input-file-type", in_type,
        "--output", str(out_path),
        "--custom-file-header", header,
    ]

    status = codegen_main(args)

    match status:
        case Exit.OK:
            return
        case Exit.ERROR:
            raise SystemExit("code generation failed")
        case Exit.KeyboardInterrupt:
            raise KeyboardInterrupt()
        case _:
            raise SystemExit(f"unknown result from code generation: {status}")


# --------- Below this point is the self-bootstrapping logic ---------


import importlib.util
import subprocess
import venv

# Name of a module to import, to test whether dependencies have been installed or not.
TEST_INSTALL_MODULE = "datamodel_code_generator"

# Directory for the virtualenv. This script is expected to run with Blender's
# build directory as its working directory.
VENV_DIR = Path("generate_datamodels_venv").resolve()

# Python executable inside the virtual environment.
VENV_PYTHON = VENV_DIR / "Scripts/python.exe" if sys.platform == "win32" else VENV_DIR / "bin/python"


def _create_virtualenv() -> None:
    """Create the virtual environment if it does not exist."""
    if VENV_DIR.exists():
        return
    print(f"Creating virtual environment at {VENV_DIR}")
    venv.create(VENV_DIR, with_pip=True)


def _install_dependencies() -> None:
    """Install required dependencies into the virtual environment."""
    print("Installing dependencies")
    # Pip doesn't like to be used as Python library, invoking it via the CLI is the best option.
    _run_command(str(VENV_PYTHON), "-m", "pip", "install", "--upgrade", "pip")
    _run_command(str(VENV_PYTHON), "-m", "pip", "install", "--upgrade", *REQUIREMENTS)


def _is_dependency_installed(package: str) -> bool:
    """Try importing a package to check if it is installed."""
    return importlib.util.find_spec(package) is not None


def _is_running_in_virtualenv() -> bool:
    """Check if the script is running inside the virtual environment."""
    return sys.prefix != sys.base_prefix  # Virtualenv modifies `sys.prefix`


def _run_command(*cmd: str) -> None:
    """Run a shell command and handle errors."""
    try:
        subprocess.run(cmd, check=True, text=True)
    except subprocess.CalledProcessError as e:
        print(f"Error running command: {' '.join(cmd)}", file=sys.stderr)
        print(f"Exit code: {e.returncode}", file=sys.stderr)
        sys.exit(e.returncode)


if __name__ == "__main__":
    _create_virtualenv()

    if not _is_running_in_virtualenv():
        print(f"Re-executing inside virtual environment at {VENV_DIR}")
        _run_command(str(VENV_PYTHON), *sys.argv)
        sys.exit()

    if not _is_dependency_installed(TEST_INSTALL_MODULE):
        _install_dependencies()

    # The virtual environment is active, so run the main script logic.
    main()



# SPDX-FileCopyrightText: 2024 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

from gdb import Type, Field


def get_basic_type(type: Type) -> Type:
    pass


def has_field(type: Type) -> bool:
    pass


def make_enum_dict(enum_type: Type) -> dict[str, int]:
    pass



# SPDX-FileCopyrightText: 2025 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

import _blendfile_header  # Not part of the Public API, allow use for testing.
import blend_render_info
import bpy
import pathlib
import sys
import unittest
import gzip
import tempfile

args = None


class BlendFileHeaderTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.testdir = args.testdir

    def setUp(self):
        self.assertTrue(self.testdir.exists(),
                        "Test dir {0} should exist".format(self.testdir))

    def test_small_bhead_8(self):
        path = self.testdir / "SmallBHead8.blend"
        with gzip.open(path, "rb") as f:
            header = _blendfile_header.BlendFileHeader(f)
            self.assertEqual(header.magic, b"BLENDER")
            self.assertEqual(header.file_format_version, 0)
            self.assertEqual(header.pointer_size, 8)
            self.assertTrue(header.is_little_endian)
            self.assertEqual(header.version, 300)

            header_struct = header.create_block_header_struct()
            self.assertIs(header_struct.type, _blendfile_header.SmallBHead8)

            buffer = f.read(header_struct.struct.size)
            block = header_struct.parse(buffer)
            self.assertEqual(block.code, b"REND")
            self.assertEqual(block.len, 72)
            self.assertEqual(block.old, 140732920740000)
            self.assertEqual(block.SDNAnr, 0)
            self.assertEqual(block.nr, 1)

        self.assertEqual(blend_render_info.read_blend_rend_chunk(path), [(1, 250, "Scene")])

    def test_large_bhead_8(self):
        path = self.testdir / "LargeBHead8.blend"
        with open(path, "rb") as f:
            header = _blendfile_header.BlendFileHeader(f)
            self.assertEqual(header.magic, b"BLENDER")
            self.assertEqual(header.file_format_version, 1)
            self.assertEqual(header.pointer_size, 8)
            self.assertTrue(header.is_little_endian)
            self.assertEqual(header.version, 500)

            header_struct = header.create_block_header_struct()
            self.assertIs(header_struct.type, _blendfile_header.LargeBHead8)

            buffer = f.read(header_struct.struct.size)
            block = header_struct.parse(buffer)
            self.assertEqual(block.code, b"REND")
            self.assertEqual(block.len, 72)
            self.assertEqual(block.old, 140737488337232)
            self.assertEqual(block.SDNAnr, 0)
            self.assertEqual(block.nr, 1)

        self.assertEqual(blend_render_info.read_blend_rend_chunk(path), [(1, 250, "Scene")])

    def test_bhead_4(self):
        path = self.testdir / "BHead4.blend"
        with gzip.open(path, "rb") as f:
            header = _blendfile_header.BlendFileHeader(f)
            self.assertEqual(header.magic, b"BLENDER")
            self.assertEqual(header.file_format_version, 0)
            self.assertEqual(header.pointer_size, 4)
            self.assertTrue(header.is_little_endian)
            self.assertEqual(header.version, 260)

            header_struct = header.create_block_header_struct()
            self.assertIs(header_struct.type, _blendfile_header.BHead4)

            buffer = f.read(header_struct.struct.size)
            block = header_struct.parse(buffer)
            self.assertEqual(block.code, b"REND")
            self.assertEqual(block.len, 32)
            self.assertEqual(block.old, 2684488)
            self.assertEqual(block.SDNAnr, 0)
            self.assertEqual(block.nr, 1)

        self.assertEqual(blend_render_info.read_blend_rend_chunk(path), [(1, 250, "Space types")])

    def test_bhead_4_big_endian(self):
        path = self.testdir / "BHead4_big_endian.blend"
        with gzip.open(path, "rb") as f:
            header = _blendfile_header.BlendFileHeader(f)
            self.assertEqual(header.magic, b"BLENDER")
            self.assertEqual(header.file_format_version, 0)
            self.assertEqual(header.pointer_size, 4)
            self.assertFalse(header.is_little_endian)
            self.assertEqual(header.version, 170)

            header_struct = header.create_block_header_struct()
            self.assertIs(header_struct.type, _blendfile_header.BHead4)

            buffer = f.read(header_struct.struct.size)
            block = header_struct.parse(buffer)
            self.assertEqual(block.code, b"REND")
            self.assertEqual(block.len, 32)
            self.assertEqual(block.old, 2147428916)
            self.assertEqual(block.SDNAnr, 0)
            self.assertEqual(block.nr, 1)

        self.assertEqual(blend_render_info.read_blend_rend_chunk(path), [(1, 150, "1")])

    def test_current(self):
        directory = tempfile.mkdtemp()
        path = pathlib.Path(directory) / "test.blend"

        bpy.ops.wm.read_factory_settings(use_empty=True)

        scene = bpy.data.scenes[0]
        scene.name = "Test Scene"
        scene.frame_start = 10
        scene.frame_end = 20
        bpy.ops.wm.save_as_mainfile(filepath=str(path), compress=False, copy=True)

        version = bpy.app.version
        version_int = version[0] * 100 + version[1]

        with open(path, "rb") as f:
            header = _blendfile_header.BlendFileHeader(f)
            self.assertEqual(header.magic, b"BLENDER")
            self.assertEqual(header.file_format_version, 1)
            self.assertEqual(header.pointer_size, 8)
            self.assertTrue(header.is_little_endian)
            self.assertEqual(header.version, version_int)

            header_struct = header.create_block_header_struct()
            self.assertIs(header_struct.type, _blendfile_header.LargeBHead8)

            buffer = f.read(header_struct.struct.size)
            block = header_struct.parse(buffer)
            self.assertEqual(block.code, b"REND")

        self.assertEqual(blend_render_info.read_blend_rend_chunk(path), [(10, 20, "Test Scene")])


def main():
    global args
    import argparse

    if '--' in sys.argv:
        argv = [sys.argv[0]] + sys.argv[sys.argv.index('--') + 1:]
    else:
        argv = sys.argv

    parser = argparse.ArgumentParser()
    parser.add_argument('--testdir', required=True, type=pathlib.Path)
    args, remaining = parser.parse_known_args(argv)

    unittest.main(argv=remaining)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2023 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Starts up a web server pointed to a local directory for the purpose of simulating online access.
With basic options for PORT/path & verbosity (so tests aren't too noisy).
"""
__all__ = (
    "HTTPServerContext",
)

import socketserver
import http.server
import threading

from typing import (
    Any,
)


class HTTPServerContext:
    __slots__ = (
        "_directory",
        "_port",
        "_http_thread",
        "_http_server",
        "_wait_tries",
        "_wait_delay",
        "_verbose",
    )

    class _TestServer(socketserver.TCPServer):
        allow_reuse_address = True

    @staticmethod
    def _is_port_in_use(port: int) -> bool:
        import socket
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            return s.connect_ex(("localhost", port)) == 0

    @staticmethod
    def _test_handler_factory(directory: str, verbose: bool = False) -> type:
        class TestHandler(http.server.SimpleHTTPRequestHandler):
            def __init__(self, *args: Any, **kwargs: Any) -> None:
                super().__init__(*args, directory=directory, **kwargs)
            # Suppress messages by overriding the function.
            if not verbose:
                def log_message(self, *_args: Any, **_kw: Any) -> None:
                    pass
        return TestHandler

    def __init__(
            self,
            directory: str,
            port: int,
            *,
            verbose: bool = False,
            wait_delay: float = 0.0,
            wait_tries: int = 0,
    ) -> None:
        self._directory = directory
        self._port = port
        self._wait_delay = wait_delay
        self._wait_tries = wait_tries
        self._verbose = verbose

        # Members `_http_thread` & `_http_server` are set when entering the context.

    def __enter__(self) -> None:

        if self._wait_tries:
            import time
            for _ in range(self._wait_tries):
                if not HTTPServerContext._is_port_in_use(self._port):
                    break

                print("Waiting...")
                time.sleep(self._wait_delay)

        http_server = HTTPServerContext._TestServer(
            ("", self._port),
            HTTPServerContext._test_handler_factory(
                self._directory,
                verbose=self._verbose,
            ),
        )

        # Use a thread so as not to block.
        http_thread = threading.Thread(target=http_server.serve_forever)
        http_thread.daemon = True
        http_thread.start()

        # pylint: disable-next=attribute-defined-outside-init
        self._http_thread = http_thread
        # pylint: disable-next=attribute-defined-outside-init
        self._http_server = http_server

    def __exit__(self, _type: Any, _value: Any, traceback: Any) -> None:
        # Needed on WIN32, otherwise exit causes an `OSError`.
        self._http_server.shutdown()

        self._http_server.server_close()
        del self._http_server
        del self._http_thread



# SPDX-FileCopyrightText: 2020-2022 Blender Authors
#
# SPDX-License-Identifier: Apache-2.0

import bpy
import os
import pprint
import unittest


class TestHelper(unittest.TestCase):

    def __init__(self, args):
        super().__init__()
        self.args = args

    @staticmethod
    def id_to_uid(id_data):
        return (type(id_data).__name__,
                id_data.name_full,
                id_data.users)

    @classmethod
    def blender_data_to_tuple(cls, bdata, pprint_name=None):
        ret = sorted(tuple((cls.id_to_uid(k), sorted(tuple(cls.id_to_uid(vv) for vv in v)))
                           for k, v in bdata.user_map().items()))
        if pprint_name is not None:
            print("\n%s:" % pprint_name)
            pprint.pprint(ret)
        return ret

    @staticmethod
    def ensure_path(path):
        if not os.path.exists(path):
            os.makedirs(path)

    def run_all_tests(self):
        for inst_attr_id in dir(self):
            if not inst_attr_id.startswith("test_"):
                continue
            inst_attr = getattr(self, inst_attr_id)
            if not callable(inst_attr):
                continue

            self.setUp()
            try:
                inst_attr()
            finally:
                self.tearDown()


class TestBlendLibLinkHelper(TestHelper):
    """
    Generate relatively complex data layout accross several blendfiles.

    Useful for testing link/append/etc., but also data relationships e.g.
    """

    def __init__(self, args):
        assert hasattr(args, "src_test_dir")
        assert hasattr(args, "output_dir")
        super().__init__(args)

    @staticmethod
    def reset_blender():
        bpy.ops.wm.read_homefile(use_empty=True, use_factory_startup=True)
        bpy.data.orphans_purge(do_recursive=True)

    def unique_blendfile_name(self, base_name):
        return base_name + self.__class__.__name__ + ".blend"

    # Utils generating common data patterns.
    # Assume that the current blendfile is cleared and empty.

    def gen_indirect_library_data_(self):
        im = bpy.data.images.load(os.path.join(self.args.src_test_dir,
                                               "imbuf_io",
                                               "reference",
                                               "jpeg-rgb-90__from__rgba08.jpg"))
        im.name = "LibMaterial"
        self.assertTrue(len(im.pixels) > 0)
        self.assertTrue(im.has_data)

        ma = bpy.data.materials.new("LibMaterial")
        ma.use_fake_user = True
        out_node = ma.node_tree.nodes.new("ShaderNodeOutputMaterial")
        bsdf_node = ma.node_tree.nodes.new("ShaderNodeBsdfPrincipled")
        ma.node_tree.links.new(bsdf_node.outputs["BSDF"], out_node.inputs["Surface"])
        teximage_node = ma.node_tree.nodes.new("ShaderNodeTexImage")
        teximage_node.image = im
        ma.node_tree.links.new(bsdf_node.inputs["Base Color"], teximage_node.outputs["Color"])

    def gen_library_data_(self):
        me = bpy.data.meshes.new("LibMesh")
        ob = bpy.data.objects.new("LibMesh", me)
        coll = bpy.data.collections.new("LibMesh")
        coll.objects.link(ob)
        bpy.context.scene.collection.children.link(coll)

    # Actual init code, generating blendfiles for various linking/appending scenarii.

    def init_lib_data_basic(self):
        self.reset_blender()

        self.gen_library_data_()

        output_dir = self.args.output_dir
        self.ensure_path(output_dir)
        # Take care to keep the name unique so multiple test jobs can run at once.
        output_lib_path = os.path.join(output_dir, self.unique_blendfile_name("blendlib_basic"))

        bpy.ops.wm.save_as_mainfile(filepath=output_lib_path, check_existing=False, compress=False)

        return output_lib_path

    def init_lib_data_animated(self):
        self.reset_blender()

        self.gen_library_data_()

        ob = bpy.data.objects[0]
        ob_ctrl = bpy.data.objects.new("LibController", None)
        coll = bpy.data.collections[0]
        coll.objects.link(ob_ctrl)

        # Add some action & driver animation to `LibMesh`.
        # Animate Y location.
        ob.location[1] = 0.0
        ob.keyframe_insert("location", index=1, frame=1)
        ob.location[1] = -5.0
        ob.keyframe_insert("location", index=1, frame=10)

        # Drive X location.
        ob_drv = ob.driver_add("location", 0)
        ob_drv.driver.type = 'AVERAGE'
        ob_drv_var = ob_drv.driver.variables.new()
        ob_drv_var.type = 'TRANSFORMS'
        ob_drv_var.targets[0].id = ob_ctrl
        ob_drv_var.targets[0].transform_type = 'LOC_X'

        # Add some action & driver animation to `LibController`.
        # Animate X location.
        ob_ctrl.location[0] = 0.0
        ob_ctrl.keyframe_insert("location", index=0, frame=1)
        ob_ctrl.location[0] = 5.0
        ob_ctrl.keyframe_insert("location", index=0, frame=10)

        output_dir = self.args.output_dir
        self.ensure_path(output_dir)
        # Take care to keep the name unique so multiple test jobs can run at once.
        output_lib_path = os.path.join(output_dir, self.unique_blendfile_name("blendlib_animated"))

        bpy.ops.wm.save_as_mainfile(filepath=output_lib_path, check_existing=False, compress=False)

        return output_lib_path

    def init_lib_data_indirect_lib(self):
        output_dir = self.args.output_dir
        self.ensure_path(output_dir)

        # Create an indirect library containing a material, and an image texture.
        self.reset_blender()

        self.gen_indirect_library_data_()

        # Take care to keep the name unique so multiple test jobs can run at once.
        output_lib_path = os.path.join(output_dir, self.unique_blendfile_name("blendlib_indirect_material"))

        bpy.ops.wm.save_as_mainfile(filepath=output_lib_path, check_existing=False, compress=False)

        # Create a main library containing object etc., and linking material from indirect library.
        self.reset_blender()

        link_dir = os.path.join(output_lib_path, "Material")
        bpy.ops.wm.link(directory=link_dir, filename="LibMaterial")
        ma = bpy.data.materials[0]

        self.gen_library_data_()

        me = bpy.data.meshes[0]
        me.materials.append(ma)

        output_dir = self.args.output_dir
        self.ensure_path(output_dir)
        # Take care to keep the name unique so multiple test jobs can run at once.
        output_lib_path = os.path.join(output_dir, self.unique_blendfile_name("blendlib_indirect_main"))

        bpy.ops.wm.save_as_mainfile(filepath=output_lib_path, check_existing=False, compress=False)

        return output_lib_path

    def init_lib_data_packed_indirect_lib(self):
        output_dir = self.args.output_dir
        self.ensure_path(output_dir)

        # Create an indirect library containing a material, and an image texture.
        self.reset_blender()

        self.gen_indirect_library_data_()

        # Take care to keep the name unique so multiple test jobs can run at once.
        output_lib_path = os.path.join(output_dir, self.unique_blendfile_name("blendlib_indirect_material"))

        bpy.ops.wm.save_as_mainfile(filepath=output_lib_path, check_existing=False, compress=False)

        # Create a main library containing object etc., and linking material from indirect library.
        self.reset_blender()

        self.gen_library_data_()

        link_dir = os.path.join(output_lib_path, "Material")
        bpy.ops.wm.link(directory=link_dir, filename="LibMaterial")

        ma = bpy.data.pack_linked_ids_hierarchy(bpy.data.materials[0])

        me = bpy.data.meshes[0]
        me.materials.append(ma)

        bpy.ops.outliner.orphans_purge()

        self.assertEqual(len(bpy.data.materials), 1)
        self.assertTrue(bpy.data.materials[0].is_linked_packed)

        self.assertEqual(len(bpy.data.images), 1)
        self.assertTrue(bpy.data.images[0].is_linked_packed)

        self.assertEqual(len(bpy.data.libraries), 2)
        self.assertFalse(bpy.data.libraries[0].is_archive)
        self.assertTrue(bpy.data.libraries[1].is_archive)
        self.assertIn(bpy.data.libraries[1].name, bpy.data.libraries[0].archive_libraries)

        output_dir = self.args.output_dir
        self.ensure_path(output_dir)
        # Take care to keep the name unique so multiple test jobs can run at once.
        output_lib_path = os.path.join(output_dir, self.unique_blendfile_name("blendlib_indirect_main"))

        bpy.ops.wm.save_as_mainfile(filepath=output_lib_path, check_existing=False, compress=False)

        return output_lib_path



# SPDX-FileCopyrightText: 2025 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later */

__all__ = (
    "main",
)

import math
import unittest
import sys
import pathlib
import numpy as np

import bpy

"""
blender -b --factory-startup --python tests/python/sculpt_paint/mask_filter_test.py -- --testdir tests/files/sculpting/
"""

args = None


class GrowMaskTest(unittest.TestCase):
    def setUp(self):
        bpy.ops.wm.open_mainfile(filepath=str(args.testdir / "partially_masked_sphere.blend"), load_ui=False)
        bpy.ops.ed.undo_push()

    def test_grow_increases_number_of_masked_vertices(self):
        mesh = bpy.context.object.data
        mask_attr = mesh.attributes['.sculpt_mask']

        num_vertices = mesh.attributes.domain_size('POINT')

        old_mask_data = np.zeros(num_vertices, dtype=np.float32)
        mask_attr.data.foreach_get('value', old_mask_data)

        bpy.ops.sculpt.mask_filter(filter_type='GROW')

        new_mask_data = np.zeros(num_vertices, dtype=np.float32)
        mask_attr.data.foreach_get('value', new_mask_data)

        self.assertGreater(np.count_nonzero(new_mask_data), np.count_nonzero(old_mask_data))


class ShrinkMaskTest(unittest.TestCase):
    def setUp(self):
        bpy.ops.wm.open_mainfile(filepath=str(args.testdir / "partially_masked_sphere.blend"), load_ui=False)
        bpy.ops.ed.undo_push()

    def test_shrink_decreases_number_of_masked_vertices(self):
        mesh = bpy.context.object.data
        mask_attr = mesh.attributes['.sculpt_mask']

        num_vertices = mesh.attributes.domain_size('POINT')

        old_mask_data = np.zeros(num_vertices, dtype=np.float32)
        mask_attr.data.foreach_get('value', old_mask_data)

        bpy.ops.sculpt.mask_filter(filter_type='SHRINK')

        new_mask_data = np.zeros(num_vertices, dtype=np.float32)
        mask_attr.data.foreach_get('value', new_mask_data)

        self.assertLess(np.count_nonzero(new_mask_data), np.count_nonzero(old_mask_data))


def main():
    global args
    import argparse

    argv = [sys.argv[0]]
    if '--' in sys.argv:
        argv += sys.argv[sys.argv.index('--') + 1:]

    parser = argparse.ArgumentParser()
    parser.add_argument('--testdir', required=True, type=pathlib.Path)

    args, remaining = parser.parse_known_args(argv)

    unittest.main(argv=remaining)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-FileCopyrightText: 2022 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

# macOS utility to remove all `rpaths` and add a new one.

__all__ = (
    "main",
)

import os
import pathlib
import re
import subprocess
import sys


# Strip version numbers from dependencies macOS notarizatiom fails
# with version symlinks.
def strip_lib_version(name):
    name = re.sub(r'(\.[0-9]+)+.dylib', '.dylib', name)
    name = re.sub(r'(\.[0-9]+)+.so', '.so', name)
    name = re.sub(r'(\.[0-9]+)+.cpython', '.cpython', name)
    return name


# Patch cmake config to match rename
def update_cmake_config(oldfile, newfile):
    for cmakefile in oldfile.parent.glob("cmake/*/*.cmake"):
        text = cmakefile.read_text()
        text = text.replace(oldfile.name, newfile.name)
        cmakefile.write_text(text)


def main():
    rpath = sys.argv[1]
    file = sys.argv[2]
    new_file = strip_lib_version(file)

    file = pathlib.Path(file)
    new_file = pathlib.Path(new_file)

    # Update CMake configuration files.
    update_cmake_config(file, new_file)

    # Remove if symbolic-link.
    if file.is_symlink():
        os.remove(file)
        sys.exit(0)

    # Find existing RPATHS and delete them one by one.
    p = subprocess.run(['otool', '-l', file], capture_output=True)
    tokens = p.stdout.split()

    for i, token in enumerate(tokens):
        if token == b'LC_RPATH':
            old_rpath = tokens[i + 4]
            subprocess.run(['install_name_tool', '-delete_rpath', old_rpath, file])

    subprocess.run(['install_name_tool', '-add_rpath', rpath, file])

    # Strip version from dependencies.
    p = subprocess.run(['otool', '-L', file], capture_output=True)
    tokens = p.stdout.split()
    for i, token in enumerate(tokens):
        token = token.decode("utf-8")
        if token.startswith("@rpath"):
            new_token = strip_lib_version(token)
            subprocess.run(['install_name_tool', '-change', token, new_token, file])

    # Strip version from library itself.
    new_id = '@rpath/' + new_file.name
    os.rename(file, new_file)
    subprocess.run(['install_name_tool', '-id', new_id, new_file])


if __name__ == "__main__":
    main()
